import json
import logging
import sys
import time
import typing
from collections import OrderedDict
from copy import copy
from pprint import pformat
from uuid import uuid4

import requests
from six import wraps
from six.moves.urllib.parse import urljoin
from six.moves.urllib.parse import urlparse

logger = logging.getLogger(__name__)
MAX_RETRIES = 5
DEFAULT_TIMEOUT = 180


class ProviderHTTPClientException(Exception):
    def __init__(self, *args, **kwargs):
        super(ProviderHTTPClientException, self).__init__(*args)
        self.response = kwargs.get('response')
        for key, value in kwargs.items():
            setattr(self, key, value)


class RetryableException(ProviderHTTPClientException):
    pass


class ProviderHTTPServiceException(ProviderHTTPClientException):
    pass


def flatten_nested_object(obj, parent_key=''):
    result = {}

    for key, value in obj.items():
        if parent_key:
            new_key = '.'.join([parent_key, key])
        else:
            new_key = key

        if isinstance(value, list):
            raise ValueError('Nesting list is not supported yet. Please change key %s' % new_key)
        if isinstance(value, dict):
            result.update(flatten_nested_object(value, new_key))
        else:
            result[new_key] = value
    return result


# noinspection PyMethodMayBeStatic
class ProviderRequestProxy(object):
    def __init__(self, *, url, headers, timeout, retries=MAX_RETRIES, host=None, retry_delay=5, **kwargs):
        self.url = url
        self.headers = headers
        self.timeout = timeout
        self.retries = retries
        self.host = host
        self.raw = kwargs.get('raw', False)
        self.is_json = kwargs.get('json', True)
        self.retry_delay = retry_delay
        self.force_dict_order = kwargs.get('force_dict_order', sys.version_info < (3, 6))

    def augment_request_arguments(self, **request_kwargs):
        self.headers.update(request_kwargs.get('headers', {}))
        url_override = request_kwargs.get('url', None)
        self.url = self.url.format(**request_kwargs)

        data = {
            'url': url_override if url_override else self.url,
            'headers': self.headers,
            'timeout': self.timeout,
        }
        if 'data' in request_kwargs:
            data['data'] = request_kwargs['data']
        if 'json' in request_kwargs:
            data['json'] = request_kwargs['json']
        if 'files' in request_kwargs:
            data['files'] = request_kwargs['files']
        if 'stream' in request_kwargs:
            data['stream'] = request_kwargs['stream']
        if 'params' in request_kwargs:
            data['params'] = request_kwargs['params']
        if 'hooks' in request_kwargs:
            data['hooks'] = request_kwargs['hooks']

        if 'json' in data and 'data' in data:
            raise ValueError('You need to use either `json` or `data` parameter, not both.')

        if ('json' in data or 'data' in data) and 'files' in data:
            # parse json or data to enable multipart
            data_payload = data.pop('json', data.pop('data', None))

            if isinstance(data_payload, str):
                data_payload = json.load(data_payload)

            if isinstance(data_payload, list):
                raise ValueError('List json is not supported with Multipart')

            data['data'] = flatten_nested_object(data_payload)

        return data

    def get(self, **request_arguments):
        return self.request_with_retry('get', **request_arguments)

    def head(self, **request_arguments):
        return self.request_with_retry('head', **request_arguments)

    def options(self, **request_arguments):
        return self.request_with_retry('options', **request_arguments)

    def post(self, **request_arguments):
        return self.request_with_retry('post', **request_arguments)

    def put(self, **request_arguments):
        return self.request_with_retry('put', **request_arguments)

    def patch(self, **request_arguments):
        return self.request_with_retry('patch', **request_arguments)

    def delete(self, **request_arguments):
        return self.request_with_retry('delete', **request_arguments)

    def format_return_data(self, response):
        # TODO: decouple it
        if response.status_code == 204 or self.raw:
            return response
        elif self.is_json:
            try:
                if self.force_dict_order:
                    return response.json(object_pairs_hook=OrderedDict)
                else:
                    return response.json()
            except ValueError as e:
                if 'JSON' not in getattr(e, 'message', ''):
                    raise
                logger.exception('Failed to decode json file with error message: %s', e)
                return {}
        else:
            return response.content

    def set_file_caret_to_start(self, **request_arguments):
        files_in_request = request_arguments.get('files')
        if files_in_request:
            try:
                for key in files_in_request.keys():
                    files_in_request[key].seek(0)
            except Exception as exc:
                logger.debug('Failed to put file pointer to the start of the file with exc: %s', exc, exc_info=exc)

    def raise_if_error_status_code(self, response: requests.Response):
        if 400 <= response.status_code < 500:
            http_error_msg = '%s Client Error: %s for url: %s %s' % \
                             (response.status_code,
                              response.reason,
                              response.url,
                              response.content)
            raise ProviderHTTPClientException(http_error_msg, response=response)
        elif response.status_code in [502, 503]:
            raise RetryableException(
                'Deployment might be in progress, or server is overloaded',
                response=response
            )
        elif 500 <= response.status_code < 600:
            raise ProviderHTTPClientException('%d Server error. %s ' % (
                response.status_code,
                response.reason
            ), response=response)

    def request_with_retry(self, method, **request_arguments):
        request_arguments = self.augment_request_arguments(**request_arguments)
        retries = 0
        retry_errors = []
        response = None

        # if we say, that there should be no retries, put retries=0
        while retries < self.retries + 1:
            try:
                self.set_file_caret_to_start(**request_arguments)
                response = getattr(requests, method)(**request_arguments)
                self.raise_if_error_status_code(response)

            except (requests.Timeout, RetryableException) as exc:
                logger.debug('Retry request %s %s', method, request_arguments['url'])
                retries += 1
                retry_errors.append(exc)
                time.sleep(self.retry_delay)
            except requests.ConnectionError as exc:
                retry_errors.append(exc)
                raise ProviderHTTPServiceException('Service is not working with error: %s' % exc,
                                                   response=response)
            else:
                break  # Exits while loop
        else:
            raise ProviderHTTPServiceException('Failed to do request after {} tries'.format(len(retry_errors)),
                                               retries=retry_errors,
                                               response=response)

        return self.format_return_data(response)


class UniqueRequestProxy(ProviderRequestProxy):
    ENABLED_METHODS = ('post', 'put')

    def __init__(self,
                 *,
                 request_uid_header_name: str,
                 request_uid_value: typing.Callable[..., str] = None,
                 **kwargs):
        """
        :param request_uid_header_name: HTTP header name that will contain the
        request UUID
        :param request_uid_value: callable that will receive the request context and
        must return the string with UUID value; if callable is not provided or
        returned None, UUID will be generated automatically for every request
        """
        super(UniqueRequestProxy, self).__init__(**kwargs)

        self.header_name = request_uid_header_name
        self.uid_value = request_uid_value

    def inject_unique_request_header(self, **request_arguments):
        uuid_value = None
        if callable(self.uid_value):
            uuid_value = self.uid_value(**request_arguments)
        if uuid_value is None:
            uuid_value = str(uuid4())

        self.headers[self.header_name] = uuid_value

    def request_with_retry(self, method, **request_arguments):
        if method in self.ENABLED_METHODS:
            self.inject_unique_request_header(**request_arguments)

        return super(UniqueRequestProxy, self).request_with_retry(
            method,
            **request_arguments
        )

    def raise_if_error_status_code(self, response: requests.Response):
        # Handle "Another request is already processing" error with retry
        if response.status_code == 424:
            try:
                json_resp = response.json()
                error_desc = json_resp['error']
            except (ValueError, TypeError, KeyError):
                error_desc = 'Another request is already processing'
            raise RetryableException(error_desc, response=response)

        return super(UniqueRequestProxy, self).raise_if_error_status_code(response)


def remote_service(path, headers=None, timeout=None, json=True, raw=False, **options):
    if headers is None:
        headers = {}
    assert isinstance(headers, dict), 'Headers should be a dict!'

    # safe_methods = ['get', 'head', 'options']
    # unsafe_methods = ['post', 'put', 'patch', 'delete']

    def wrapper(func):
        @wraps(func)
        def inner(self, *args, **kwargs):
            if not isinstance(self, ReboticsBaseProvider):
                raise TypeError('This decorator should be used only for subclasses Rebotics Base Providers')

            provider = copy(self)

            authentication_headers = provider.get_provider_headers()
            headers.update(authentication_headers)

            url = provider.build_url(path)
            raw_override = kwargs.get('raw')
            retries = options.pop('retries', provider.retries)
            proxy_opts = copy(provider.proxy_config)
            proxy_opts.update(options)

            session = provider.proxy_class(
                url=url,
                headers=headers,
                timeout=timeout or provider.timeout,
                retries=retries,
                json=json,
                raw=raw_override or raw,
                host=provider.host,
                force_dict_order=provider.force_dict_order,
                **proxy_opts
            )

            provider.session = session
            self.session = session

            setattr(func, 'provider', provider)
            setattr(func, 'url', url)

            return func(provider, *args, **kwargs)

        return inner

    return wrapper


# noinspection PyMethodMayBeStatic
class ReboticsBaseProvider(object):
    retries = MAX_RETRIES
    timeout = DEFAULT_TIMEOUT
    proxy_class = ProviderRequestProxy
    proxy_config = {}  # additional parameters that will be passed to proxy_cls' init

    def __init__(self, host, **kwargs):
        self.host = host
        self.domain = urlparse(self.host).netloc
        self.headers = kwargs.get('headers', {})
        self.data = kwargs.get('data', {})
        self.session = None
        self.requests = requests

        if 'retries' in kwargs:
            self.retries = kwargs['retries']

        if 'timeout' in kwargs:
            self.timeout = kwargs['timeout']

        if 'token' in kwargs:
            self.set_token(kwargs['token'])

        self.role = kwargs.get('role', '')
        self.force_dict_order = kwargs.get('force_dict_order', sys.version_info < (3, 6))

    def get_provider_headers(self):
        return self.headers if self.headers else {}

    def set_header(self, key, value):
        self.headers[key] = value

    def set_token(self, token):
        self.headers['Authorization'] = 'Token %s' % token

    def build_url(self, path):
        return urljoin(self.host, path)

    @remote_service('/ping/', json=False)
    def ping(self, **kwargs):
        return self.session.get()

    @remote_service('/notifications/setWebhook/')
    def set_webhook(self, url, token=None):
        """Set webhook used by rebotics_sdk.hooks and rebotics_sdk.notification apps"""
        data = {
            'url': url
        }
        if token is not None:
            data['auth_token'] = token
        return self.session.post(data=data)

    def __repr__(self):
        return '<{} {}>'.format(self.__class__.__name__, pformat({
            'host': self.host,
            'headers': self.headers,
            'role': self.role
        }))

    def set_retailer_identifier(self, retailer_codename, retailer_secret_key):
        self.headers['x-retailer-codename'] = retailer_codename
        self.headers['x-retailer-secret-key'] = retailer_secret_key

    def is_retailer_identifier_used(self):
        header_keys = self.headers.keys()
        return ('x-retailer-codename' in header_keys) and ('x-retailer-secret-key' in header_keys)


class PageResult(dict):
    def __init__(self, json_data):
        """
        Returns paged result from rest framework
        :param dict json_data:
        """
        self.json_data = json_data
        super(PageResult, self).__init__(json_data)

    def __len__(self):
        return self.json_data['count']

    def __iter__(self):
        for item in self.json_data['results']:
            yield item

    def __getitem__(self, item):
        return self.json_data['results'][item]

    def __repr__(self):
        return "PageResult of {}".format(len(self))

    @property
    def pdf_download_url(self):
        return self.json_data.get('pdf_url')

    @property
    def next_page(self):
        return self.json_data['next']


class APIPageIterator:
    """
    Basic usage:
        iterator = APIPageIterator(provider.get_products_list, 1)
        for data in iterator:
            # do whatever with the data
            pass
    """
    def __init__(self, method, initial_page=1, **filter_params):
        self.method = method
        self.filter_params = filter_params  # page size can be passed as an optional value
        self.page = initial_page

    def __iter__(self):
        while True:
            page_result = self.method(page=self.page, **self.filter_params)

            yield from page_result

            if not page_result.next_page:
                break

            self.page += 1


def required_model_params(params=None):
    if params is None:
        params = ['model_path', 'index_path', 'meta_path']

    def outer(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            model_params = {
                key: kwargs[key] for key in params
            }
            kwargs['model_params'] = {
                'model_path': model_params['model_path'],
                'model_index_path': model_params['index_path'],
                'model_meta_path': model_params['meta_path']
            }
            result = func(*args, **kwargs)
            return result

        return wrapper

    if callable(params):
        return outer(params)

    return outer



import random

RANDOM_STRING_SYMBOLS = 'abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789'


def get_random_string(length: int, allowed_symbols: str = RANDOM_STRING_SYMBOLS) -> str:
    """
    Return a random string with defined length in the same way as
    django.utils.crypto.get_random_string does.
    But this method is not cryptographically strong, it should not be used anywhere
    outside the tests.
    """
    return ''.join(random.choice(allowed_symbols) for _ in range(length))



from io import BytesIO
import json
from mock import patch
from unittest.case import TestCase
from urllib.parse import urljoin
from uuid import uuid4

import requests
import requests_mock

from rebotics_sdk.providers.base import (
    UniqueRequestProxy, ReboticsBaseProvider, remote_service, RetryableException
)
from tests.utils import get_random_string


def mirror_request_data_to_json(request: requests.Request, _context) -> dict:
    """
    Mirrors the incoming request method and headers to the response for test purposes.
    Some standard headers are ignored to not litter up tests
    """
    ignored_headers = ('User-Agent', 'Accept-Encoding', 'Accept', 'Connection',
                       'Content-Length', 'Content-Type')
    return {
        'method': request.method,
        'headers': {
            header: value
            for header, value in request.headers.items()
            if header not in ignored_headers
        }
    }


# To avoid hitting real APIs
@requests_mock.Mocker()
class UniqueRequestProxyIntegrationTest(TestCase):
    """
    Integration tests for UniqueRequestProxy
    """
    class TestProvider(ReboticsBaseProvider):
        TEST_HEADER_NAME = 'x-reb-test-req-uuid'

        proxy_class = UniqueRequestProxy
        proxy_config = {
            'request_uid_header_name': TEST_HEADER_NAME
        }

        TEST_METHOD_URI = '/dummy/'

        @remote_service(TEST_METHOD_URI)
        def dummy_call(self, method: str):
            fn = getattr(self.session, method.lower())
            return fn()

        @remote_service(TEST_METHOD_URI,
                        request_uid_value=lambda **kwargs: kwargs['json']['some_uuid'])
        def dummy_call_external_uuid(self, method: str, some_uuid: str):
            fn = getattr(self.session, method.lower())
            return fn(json={'some_uuid': some_uuid})

    def setUp(self):
        self.host_url = 'https://' + get_random_string(6) + '/'
        self.provider = self.TestProvider(self.host_url)
        self.api_url = urljoin(self.host_url, self.provider.TEST_METHOD_URI)

    # To have better control over generated UUID
    @patch('rebotics_sdk.providers.base.uuid4')
    def test_post_and_put_contain_unique_request_header(
            self, mock_requests, mock_uuid4):
        """
        Checks that methods that are based on PUT and POST requests will produce a
        request containing deduplication header
        """
        test_uuid4 = str(uuid4())
        mock_uuid4.return_value = test_uuid4

        methods = ('PUT', 'POST')
        for method in methods:
            with self.subTest(msg=f'{method}'):
                mock_requests.register_uri(
                    method,
                    self.api_url,
                    json=mirror_request_data_to_json
                )

                result = self.provider.dummy_call(method)

                self.assertEqual(result, {
                    'method': method,
                    'headers': {
                        self.provider.TEST_HEADER_NAME: test_uuid4
                    }
                })

    def test_other_methods_work_normally(self, mock_requests):
        """
        Checks that methods that are based on OPTIONS, HEAD, GET, PATCH and DELETE
        requests will produce a request without deduplication header as usual
        """
        methods = ('OPTIONS', 'HEAD', 'GET', 'PATCH', 'DELETE')
        for method in methods:
            with self.subTest(msg=f'{method}'):
                mock_requests.register_uri(
                    method,
                    self.api_url,
                    json=mirror_request_data_to_json
                )

                result = self.provider.dummy_call(method)

                self.assertEqual(result, {'method': method, 'headers': {}})

    def test_external_uuid_is_taken_to_header(self, mock_requests):
        """
        Checks that if method is configured to use external UUID, and it's specified,
        then this external value will be picked to the header
        """
        test_uuid4 = str(uuid4())
        method = 'POST'
        mock_requests.register_uri(
            method,
            self.api_url,
            json=mirror_request_data_to_json
        )

        result = self.provider.dummy_call_external_uuid(
            method,
            some_uuid=test_uuid4
        )

        self.assertEqual(result, {
            'method': method,
            'headers': {
                self.provider.TEST_HEADER_NAME: test_uuid4
            }
        })


class UniqueRequestProxyBaseTestCase(TestCase):
    """
    Base case for unit-testing of UniqueRequestProxy
    """
    def setUp(self):
        self.url = 'https://' + get_random_string(6) + '/'
        self.header_name = 'x-test-request-id'
        self.proxy = UniqueRequestProxy(
            url=self.url,
            headers={},
            timeout=5,
            request_uid_header_name=self.header_name
        )


class TestUniqueRequestProxyRaiseIfErrorStatusCode(UniqueRequestProxyBaseTestCase):
    """
    Unit-tests for `UniqueRequestProxy.raise_if_error_status_code`
    """
    DEFAULT_ERROR_TEXT = 'Another request is already processing'

    def test_424_with_provided_error_text(self):
        """
        Checks that error with code 424, that is usually thrown from Rebotics service,
        is considered as retryable
        """
        test_error_description = 'Request is processing: %s' % get_random_string(6)
        resp = requests.Response()
        resp.raw = BytesIO(bytes(
            json.dumps({'error': test_error_description}),
            encoding='utf8'
        ))
        resp.status_code = 424

        with self.assertRaises(RetryableException) as context:
            self.proxy.raise_if_error_status_code(resp)

        self.assertEqual(context.exception.args[0], test_error_description)

    def test_424_with_no_error_text(self):
        """
        Checks that error with code 424 but without 'error' field is considered
        as retryable
        """
        resp = requests.Response()
        resp.raw = BytesIO(bytes(
            json.dumps({}),
            encoding='utf8'
        ))
        resp.status_code = 424

        with self.assertRaises(RetryableException) as context:
            self.proxy.raise_if_error_status_code(resp)

        self.assertEqual(context.exception.args[0], self.DEFAULT_ERROR_TEXT)

    def test_424_no_json_resp(self):
        """
        Checks that error with code 424 but without json body is considered
        as retryable
        """
        resp = requests.Response()
        resp.raw = BytesIO(bytes(
            'Hi there',
            encoding='utf8'
        ))
        resp.status_code = 424

        with self.assertRaises(RetryableException) as context:
            self.proxy.raise_if_error_status_code(resp)

        self.assertEqual(context.exception.args[0], self.DEFAULT_ERROR_TEXT)



import io
import typing

from .base import ReboticsBaseProvider, remote_service, PageResult


class HawkeyeProvider(ReboticsBaseProvider):

    @remote_service('/api-token-auth/', raw=True)
    def token_auth(self, username, password, **kwargs):
        response = self.session.post(data={
            'username': username,
            'password': password
        })
        self.set_token(response.json()['token'])
        return response

    @remote_service('/api/v1/camera/heartbeats/')
    def save_camera_heartbeat(self, shelf_camera, battery_status, wifi_signal_strength, current_time):
        return self.session.post(json={
            "shelf_camera": shelf_camera,
            "battery_status": battery_status,
            "wifi_signal_strength": wifi_signal_strength,
            "current_time": current_time
        })

    @remote_service('/api/v1/camera/camera-actions/')
    def save_camera_action(self, action_type, status_type, payload, shelf_camera):
        return self.session.post(json={
            "action_type": action_type,
            "status_type": status_type,
            "payload": payload,
            "shelf_camera": shelf_camera
        })

    @remote_service('/api/v1/camera/camera-actions/')
    def get_camera_actions(self):
        return self.session.get()

    @remote_service('/api/v1/camera/retailer/')
    def save_retailer(self, codename, token):
        return self.session.post(json={
            "retailer": codename,
            "token": token
        })

    @remote_service('/api/v1/camera/fixtures/')
    def save_fixture(self, retailer, store_id, aisle, section):
        return self.session.post(
            json={
                "store_id": store_id,
                "aisle": aisle,
                "section": section,
                "retailer": retailer
            }
        )

    @remote_service('/api/v1/camera/fixtures/{id}')
    def delete_fixture(self, pk):
        return self.session.delete(id=pk)

    @remote_service('/api/v1/camera/fixtures/')
    def get_fixtures(self):
        return self.session.get()

    @remote_service('/api/v1/camera/shelf-cameras/')
    def create_shelf_camera(self, camera_id, added_by, fixture=None):
        data = {
            "camera_id": camera_id,
            "added_by": added_by,
        }
        if fixture is not None:
            data["fixture"] = fixture
        return self.session.post(json=data)

    @remote_service('/api/v1/camera/shelf-cameras/')
    def get_shelf_cameras(self):
        return self.session.get()

    @remote_service('/api/v1/camera/shelf-cameras/{id}/')
    def get_shelf_camera(self, id_):
        return self.session.get(id=id_)

    @remote_service('/api/v1/camera/shelf-cameras/{id}/')
    def update_shelf_camera(
        self,
        id_,
        camera_id: str = None,
        added_by: int = None,
        fixture: int = None,
        perspective_warp: typing.List[typing.List[int]] = None,
        force_null=False
    ):
        data_to_update = {
            "camera_id": camera_id,
            "added_by": added_by,
            "fixture": fixture,
            "perspective_warp": perspective_warp
        }

        return self.session.patch(
            id=id_,
            json={k: v for k, v in data_to_update.items() if not force_null and v is not None}
        )

    @remote_service('/api/v1/fetcher/{camera_id}/')
    def save_capture(self, camera, file_key, bucket_name):
        return self.session.post(
            camera_id=camera,
            json={
                "file_key": file_key,
                "bucket_name": bucket_name
            })

    @remote_service('/api/v1/camera/mac-address/search/')
    def search_camera_by_mac(self, mac_address):
        assert ":" in mac_address, "Mac address should be separated by : (colon)"
        return self.session.post(json={
            'mac_address': mac_address
        })

    @remote_service('/api/v1/camera/mac-address/')
    def create_mac_address_link(self, mac_address, vendor_secret, rebotics_secret):
        assert ":" in mac_address, "Mac address should be separated by : (colon)"
        return self.session.post(json={
            'mac_address': mac_address,
            'vendor_secret': vendor_secret,
            'rebotics_secret': rebotics_secret,
        })

    @remote_service('/api/v1/camera/mac-address/{id}/')
    def update_mac_address_link(self, id_, mac_address, vendor_secret, rebotics_secret):
        assert ":" in mac_address, "Mac address should be separated by : (colon)"
        assert isinstance(id_, int), "Link ID should be integer, use search by mac-address to determine Link ID"
        return self.session.patch(id=id_, json={
            'mac_address': mac_address,
            'vendor_secret': vendor_secret,
            'rebotics_secret': rebotics_secret,
        })

    @remote_service('/api/v1/camera/mac-address/{id}/', raw=True)
    def delete_mac_address_link(self, id_):
        response = self.session.delete(id=id_)
        response.raise_for_status()

    @remote_service('/api/v1/camera/shelf-cameras/{shelf_camera_id}/captures/{capture_id}/warped/',
                    raw=True, stream=True, allow_redirects=True)
    def get_warped_image(self, shelf_camera_id, capture_id, polygon):
        """
        Return warped image with given values.
        Return value is io.BytesIO
        """
        assert len(polygon) == 4, "There should be 4 points"
        assert all(len(coordinate) == 2 for coordinate in polygon), "They should be in format [x, y]"
        params = {
            'polygon': ','.join(
                str(point)
                for coordinate in polygon
                for point in coordinate
            )
        }
        response = self.session.get(
            params=params,
            shelf_camera_id=shelf_camera_id,
            capture_id=capture_id,
            stream=True
        )
        response.raise_for_status()
        fp = io.BytesIO()
        for chunk in response.iter_content(chunk_size=1024):
            fp.write(chunk)
        fp.seek(0)
        return fp

    @remote_service('/api/v1/camera/shelf-cameras/{shelf_camera_id}/captures/')
    def get_shelf_camera_captures(self, shelf_camera_id, page=None):
        """Get shelf camera captures"""
        if page is None:
            page = 1
        params = {
            'page': page,
        }
        return PageResult(
            self.session.get(shelf_camera_id=shelf_camera_id, params=params)
        )



from unittest import TestCase
from mock import patch, MagicMock
from rebotics_sdk.providers.base import ProviderRequestProxy
from rebotics_sdk.providers.retailer import RetailerProvider


class RetailerProviderTestCase(TestCase):
    def setUp(self) -> None:
        self.provider = RetailerProvider('https://example.com')

    @patch.object(ProviderRequestProxy, 'post')
    def test_export_to_cvat(self, mock_func: MagicMock):
        scan_ids = [4, 3, 4, 2, 1, 2, 1, 3]
        image_quality = 70
        segment_size = 30
        workspace = 'Workspace'
        self.provider.export_to_cvat(scan_ids, image_quality, segment_size, workspace)
        mock_func.assert_called_once_with(json={
            'scans': [1, 2, 3, 4],
            'image_quality': image_quality,
            'segment_size': segment_size,
            'workspace': workspace,
        })



import hashlib
import re
from functools import partial

from six import BytesIO


def hash_file(file, block_size=65536):
    hash_ = hashlib.md5()
    for buf in iter(partial(file.read, block_size), b''):
        hash_.update(buf)

    return hash_.hexdigest()


def accept_file(self, file_):
    if isinstance(file_, str):
        if '://' in file_:
            # URL supplied
            response = self.requests.get(url=file_, timeout=(30, 60))
            file_io = BytesIO(response.content)
        else:
            file_io = open(file_, 'rb')
    else:
        file_io = file_
        assert hasattr(file_io, 'read'), 'Non file like object was supplied'

    return file_io


def is_valid_url(url):
    regex = re.compile(
        r'^(?:http|ftp)s?://'  # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
        r'localhost|'  # localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
        r'(?::\d+)?'  # optional port
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)

    return re.match(regex, "http://www.example.com") is not None  # True


def get_id_list_from_ranges(id_list_str):
    id_list = []
    ranges = id_list_str.split(',')
    for number_range in ranges:
        range_with_dash = number_range.split('-')
        if len(range_with_dash) > 1:
            first_range_number = int(range_with_dash[0])
            second_range_number = int(range_with_dash[1])
            if first_range_number < second_range_number:
                id_list.extend(list(range(first_range_number, second_range_number + 1)))
            else:
                id_list.extend(list(range(second_range_number, first_range_number + 1)))
        else:
            id_list.append(number_range.strip())
    return [int(i) for i in id_list]



import requests_mock

from rebotics_sdk.providers import FVMProvider


def test_sending_to_fvm():
    with requests_mock.Mocker() as m:
        codename = 'retailer_x'
        secret_key = 'secret_key_x'
        products_count = 10
        file_url = 'file_url'
        headers = {
            'HTTP_X_RETAILER_CODENAME': codename,
            'HTTP_X_RETAILER_SECRET_KEY': secret_key,
        }
        json = {
            "retailer_codename": codename,
            "products_count": products_count,
            "file": file_url
        }
        m.post(url='http://test.com/api/synchronization/', status_code=201, headers=headers, json=json)
        provider = FVMProvider('http://test.com')
        provider.set_retailer_identifier(codename=codename, secret_key=secret_key)

        response = provider.send_products_file_to_fvm(codename, products_count, file_url)
        assert response is not None



from unittest import TestCase

import requests_mock

from rebotics_sdk.providers.cvat import CvatProvider


class CvatTestCase(TestCase):
    def setUp(self) -> None:
        self.provider = CvatProvider('http://test.com')
        self.provider.set_retailer_identifier('retailer', 'secret-key')

    def test_import_training_data(self):
        with requests_mock.Mocker() as m:
            m.post('http://test.com/api/retailer_import', json={'result': 'ok'})
            response = self.provider.start_retailer_import(data={'image': 'url'})
            assert response['result'] == 'ok'
            headers = m.last_request.headers
            assert headers['X-Retailer-Codename'] == 'retailer'
            assert headers['X-Retailer-Secret-Key'] == 'secret-key'

    def test_check_import_progress(self):
        with requests_mock.Mocker() as m:
            m.get('http://test.com/api/retailer_import/1', json={'result': 'ok'})
            response = self.provider.check_import_progress(task_id=1)
            assert response['result'] == 'ok'
            headers = m.last_request.headers
            assert headers['X-Retailer-Codename'] == 'retailer'
            assert headers['X-Retailer-Secret-Key'] == 'secret-key'



from unittest import TestCase
from mock import patch, MagicMock
from click.testing import CliRunner

from rebotics_sdk.providers.retailer import RetailerProvider
from rebotics_sdk.cli.utils import ReboticsScriptsConfiguration
from rebotics_sdk.cli.retailer import api


class RetailerProviderTestCase(TestCase):
    def setUp(self) -> None:
        self.runner = CliRunner()

    @patch('rebotics_sdk.cli.retailer.process_role')
    @patch.object(ReboticsScriptsConfiguration, 'get_provider')
    @patch.object(RetailerProvider, 'export_to_cvat')
    def test_export_to_cvat(self, export_mock: MagicMock, provider_mock: MagicMock, *args):
        scan_ids = [1, 2, 3]
        image_quality = 70
        segment_size = 30
        workspace = 'Workspace'
        command = 'export-to-cvat'
        provider_mock.return_value = RetailerProvider('https://example.com')

        self.runner.invoke(api, [
            command,
            '-q', image_quality,
            '-s', segment_size,
            '-w', workspace,
            *[str(i) for i in scan_ids]
        ])
        export_mock.assert_called_once_with(
            tuple(scan_ids),
            image_quality,
            segment_size,
            workspace,
        )



from collections import OrderedDict

import requests
import six
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor
from tqdm import tqdm


class ProgressBar(tqdm):

    def update_to(self, n):
        """
        identical to update, except `n` should be current value and not delta.
        """
        self.update(n - self.n)


PRESERVED_ORDER_V1 = ("policy", "AWSAccessKeyId", "key", "signature")
PRESERVED_SIGNATURE_V4 = (
    "key", "success_action_status",
    "policy", "x-amz-credential", "x-amz-algorithm",
    "x-amz-date", "x-amz-signature"
)
HTTP_TIMEOUTS = (30, 60)  # first is connection, second is for reading


def _order_upload_keys(fields, keys):
    output_fields = OrderedDict()
    for key in keys:
        if fields.get(key):
            output_fields[key] = fields[key]
    return output_fields


def format_request_body(fields):
    if len(fields.keys()) == 0:
        return fields
    elif set(fields.keys()).issubset(set(PRESERVED_ORDER_V1)):
        return _order_upload_keys(fields, PRESERVED_ORDER_V1)
    elif set(fields.keys()).issubset(set(PRESERVED_SIGNATURE_V4)):
        return _order_upload_keys(fields, PRESERVED_SIGNATURE_V4)
    else:
        return fields


def upload(destination, file, progress_bar=False, filename='features_backup.rcdb', ):
    """

    :param dict destination: an S3 presigned upload object
    :param file: File-like object that has read
    :param progress_bar: display progress bar or not
    :param str filename: name of file when doing the upload
    :return:
    """
    url = destination["url"]

    provider = destination.get('provider', 'aws-s3')
    if provider == 'aws-s3':
        fields = destination.get('fields', None)
        if fields is None:
            fields = dict()

        # apply sorting required for the AWS
        fields = format_request_body(fields)

        fields["file"] = (filename, file)
        encoder = MultipartEncoder(fields=fields)

        headers = {"Content-Type": encoder.content_type}
        headers.update(destination.get('headers', {}))

        if not progress_bar:
            return requests.post(
                url,
                data=encoder,
                headers=headers,
                timeout=HTTP_TIMEOUTS,
            )

        with ProgressBar(total=encoder.len, unit="bytes", unit_scale=True, leave=False) as bar:
            monitor = MultipartEncoderMonitor(
                encoder,
                lambda monitor: bar.update_to(monitor.bytes_read)
            )
            return requests.post(
                url,
                data=monitor,
                headers=headers,
                timeout=HTTP_TIMEOUTS,
            )
    elif provider == 'azure-blob' or provider == 'aws-s3-put':
        return requests.put(
            url,
            data=file,
            headers=destination.get('headers'),
            timeout=HTTP_TIMEOUTS,
        )


def download(source, destination=None, progress_bar=False):
    r = requests.get(
        source,
        stream=True,
        allow_redirects=True,
        timeout=HTTP_TIMEOUTS,
    )
    r.raise_for_status()
    return process_file_response(r, destination, progress_bar)


def process_file_response(response, destination, progress_bar=False):
    is_file = False

    if destination is None:
        fp = six.BytesIO()
    elif hasattr(destination, 'write') and hasattr(destination, 'read'):
        # it is a file-like object
        fp = destination
    else:
        fp = open(destination, 'wb')
        is_file = True

    content_length = response.headers.get('content-length')
    if content_length:
        total_length = int(content_length)
    else:
        total_length = 0
        progress_bar = False

    chunk, chunk_size = 1, 1024

    for chunk in tqdm(
        response.iter_content(chunk_size), leave=False, disable=not progress_bar,
        unit='iB', unit_scale=True, miniters=1, unit_divisor=1024, total=total_length,
    ):
        fp.write(chunk)

    if is_file:
        fp.close()
        return destination
    return fp



import json
import os
from concurrent.futures import ThreadPoolExecutor, as_completed
from multiprocessing import Pool
from urllib.parse import urlparse, unquote

import click
import six
import tqdm

from rebotics_sdk.cli.renderers import format_processing_action_output, format_full_table
from rebotics_sdk.utils import download_file
from ..advanced import remote_loaders
from ..providers import RetailerProvider
from ..utils import mkdir_p

if six.PY2:
    FileNotFoundError = IOError

app_dir = click.get_app_dir('rebotics-scripts')
mkdir_p(app_dir)


class DumpableConfiguration(object):
    def __init__(self, path):
        self.path = path

    @property
    def filepath(self):
        return os.path.expanduser(self.path)

    @property
    def config(self):
        try:
            with open(self.filepath, 'r') as config_buffer:
                return json.load(config_buffer)
        except FileNotFoundError:
            self.config = {}
            return {}
        except (json.JSONDecodeError,):
            return {}

    @config.setter
    def config(self, value):
        with open(self.filepath, 'w') as config_buffer:
            json.dump(value, config_buffer, indent=2)

    def update_configuration(self, key, **configuration):
        current_configuration = self.config
        if key not in current_configuration:
            current_configuration[key] = configuration
        else:
            current_configuration[key].update(configuration)
        self.config = current_configuration


class ReboticsScriptsConfiguration(DumpableConfiguration):
    def __init__(self, path, provider_class=RetailerProvider):
        super(ReboticsScriptsConfiguration, self).__init__(path)
        self.provider_class = provider_class

    def get_provider(self, key):
        config = self.config.get(key, None)
        if config is None:
            return None

        provider_kwargs = {
            'host': config['host'],
            'role': key,
        }
        if 'token' in config:
            provider_kwargs['token'] = config['token']

        provider = self.provider_class(**provider_kwargs)
        return provider


class ReboticsCLIContext(object):
    configuration_class = ReboticsScriptsConfiguration

    def __init__(self, command, role, format, verbose, config_path, provider_class):
        self.command = command
        self.role = role
        self.format = format
        self.verbose = verbose
        self.config_path = config_path
        self.provider_class = provider_class
        self.verbose_log(f"Config path: {self.config_path}")
        self.config_provider = self.configuration_class(self.config_path, self.provider_class)
        self.config = self.config_provider.config

    @property
    def provider(self):
        provider = self.config_provider.get_provider(self.role)

        if provider is None:
            raise click.ClickException(f'Role {self.role} is not configured.\n'
                                       f'Run: \n{self.command} -r {self.role} configure')
        return provider

    def update_configuration(self, **configuration):
        self.config_provider.update_configuration(self.role, **configuration)

    def format_result(self, items, max_column_length=30, keys_to_skip=None, force_format=None):
        force_format = force_format if force_format else self.format
        if force_format == 'json':
            click.echo(json.dumps(items, indent=2))
        elif force_format == 'id':
            click.echo(" ".join([str(item.get('id')) for item in items]))
        else:
            format_full_table(items, max_column_length=max_column_length, keys_to_skip=keys_to_skip)

    def verbose_log(self, message):
        if self.verbose:
            click.echo(message)


pass_rebotics_context = click.make_pass_decorator(ReboticsCLIContext, True)


class GroupWithOptionalArgument(click.Group):
    def parse_args(self, ctx, args):
        if args:
            if args[0] in self.commands:
                if len(args) == 1 or args[1] not in self.commands:
                    args.insert(0, '')
        super(GroupWithOptionalArgument, self).parse_args(ctx, args)


states = DumpableConfiguration(os.path.join(app_dir, 'command_state.json'))


def read_saved_role(command_name):
    roles = states.config.get('roles')
    if roles is None:
        return None
    role = roles.get(command_name)
    if role is not None:
        click.echo('Using previous role: {}'.format(role), err=True)
    return role


def process_role(ctx, role, command_name):
    if not role:
        if ctx.invoked_subcommand != 'roles':
            raise click.ClickException(
                'You have not specified role to use. User `roles` sub command to see what roles are available'
            )
    else:
        states.update_configuration('roles', **{command_name: role})


def task_runner(ctx, task_func, ids, concurrency, **kwargs):
    task_arguments = []

    for id_ in ids:
        arguments = {
            'ctx': ctx,
            'id': id_,
        }
        arguments.update(kwargs)
        task_arguments.append(arguments)

    pool = Pool(concurrency)
    data_list = pool.map(task_func, task_arguments)

    format_processing_action_output(data_list, ctx.format)


def download_file_from_dict(d):
    ctx = d['ctx']
    if ctx.verbose:
        click.echo('>> Downloading file into %s' % d['filepath'], err=True)
    result = download_file(d['url'], d['filepath'])
    click.echo('<< Downloaded file into %s' % d['filepath'], err=True)
    return result


class UnrecognizedInputTypeByExtension(Exception):
    pass


def guess_input_type(ext):
    if ext.startswith('.'):
        ext = ext.strip('.')
    if ext in [
        'jpeg', 'jpg', 'png',
    ]:
        return 'image'
    elif ext in [
        'mp4', 'mov', 'avi'
    ]:
        return 'video'
    else:
        raise UnrecognizedInputTypeByExtension('File with extension %s is given' % ext)


def downloads_with_threads(files, concurrency):
    results = []
    with ThreadPoolExecutor(concurrency) as executor:
        futures = [
            executor.submit(remote_loaders.download, file[0], file[1])
            for file in files
        ]
        with click.progressbar(length=len(futures), label='Downloading files:') as bar:
            for future in as_completed(futures):
                results.append(future.result())
                bar.update(1)
    return results


def refresh_with_threads(refresh_url, per_image, concurrency):
    results = []
    with ThreadPoolExecutor(concurrency) as executor:
        futures = [
            executor.submit(refresh_url, img_dict['remote_url'])
            for img_dict in per_image
        ]
        with click.progressbar(length=len(futures), label='Refreshing urls:') as bar:
            for future in as_completed(futures):
                results.append(future.result()['url'])
                bar.update(1)
    return results


def run_with_processes(invoked, iterable, concurrency):
    with Pool(concurrency) as pool:
        for _ in tqdm.tqdm(
            pool.starmap(invoked, iterable)
        ):
            pass
        pool.close()
        pool.join()


def get_segmentation(segmentation_filepath):
    with open(segmentation_filepath, 'r') as fd:
        segmentation_file = json.load(fd)
    return segmentation_file['per_image']


def get_segmentation_mode(segmentation):
    mode = 'items'
    if segmentation[0].get('remote_url') is not None:
        mode = 'remote_url'
    return mode


def refresh_url(refresh_url, segmentation_per_image):
    concurrency = len(segmentation_per_image)
    refreshed_urls = refresh_with_threads(refresh_url, segmentation_per_image, concurrency)
    return refreshed_urls


def save_masks(root_folder, urls):
    path_to_save_masks = root_folder / 'all_masks'
    path_to_save_masks.mkdir(parents=True, exist_ok=True)

    list_with_downloaded_mask = [
        [url, path_to_save_masks / str(urlparse(unquote(url)).path).lstrip('/').split('/')[-1]]
        for url in urls
    ]

    downloads_with_threads(list_with_downloaded_mask, concurrency=len(list_with_downloaded_mask))



import os
import pathlib

from rebotics_sdk.cli.utils import run_with_processes


def draw_annotated_drawing(root_folder, segmentation_per_image, mode, concurrency):
    calls = [f for f in os.scandir(root_folder) if f.is_dir()]

    for call in calls:
        if mode == 'items':
            result_folder = pathlib.Path(call.path) / 'annotated_img'
            result_folder.mkdir(parents=True, exist_ok=True)
            call_frames = get_call_frames(call, segmentation_per_image, mode, result_folder)
            run_with_processes(invoked=draw_bboxes, iterable=call_frames, concurrency=concurrency)
        if mode == 'remote_url':
            result_folder = pathlib.Path(call.path)
            call_frames = get_call_frames(call, segmentation_per_image, mode, result_folder)
            run_with_processes(invoked=draw_mask, iterable=call_frames, concurrency=concurrency)


def get_call_frames(call, image_segmentation, mode, result_folder):
    call_frames = [
        (frame.name, frame.path, image_segmentation, result_folder, mode) for f in os.scandir(call)
        if f.is_dir() and f.name == 'frames'
        for frame in os.scandir(f) if frame.is_file()
    ]
    return call_frames


def draw_bboxes(image_name, image_path, per_image, result_folder, mode):
    import cv2

    GREEN = (0, 255, 0,)
    RED = (0, 0, 255)

    cv2_image = cv2.imread(str(image_path))
    figure_to_draw = get_parts(mode, image_name, per_image, result_folder)

    if (cv2_image is not None) and figure_to_draw:
        for items in figure_to_draw:
            color = GREEN
            if items:
                x1 = items['x1']
                y1 = items['y1']
                x2 = items['x2']
                y2 = items['y2']
                category = items['category']
                if category != 'item':
                    color = RED
                cv2.rectangle(cv2_image, (int(x1), int(y1)), (int(x2), int(y2)), color=color, thickness=2)

        local_drawn_image_path = str(result_folder / image_name)
        cv2.imwrite(str(local_drawn_image_path), cv2_image)


def draw_mask(image_name, image_path, per_image, result_folder, mode):
    import cv2

    cv2_image = cv2.imread(str(image_path))
    figure_to_draw = get_parts(mode, image_name, per_image, result_folder)

    if (cv2_image is not None) and figure_to_draw:
        id_to_color = {
            1: [255, 0, 0],
            2: [0, 0, 255]
        }
        annotated_mask = pathlib.Path(result_folder) / 'annotated_mask'
        annotated_mask.mkdir(parents=True, exist_ok=True)

        mask = cv2.imread(str(figure_to_draw))

        for idx, color in id_to_color.items():
            mask[(mask == idx).all(-1)] = color

        result_image = cv2.addWeighted(mask, 0.6, cv2_image, 0.5, 0)

        local_drawn_image_path = str(annotated_mask / image_name)
        cv2.imwrite(str(local_drawn_image_path), result_image)


def get_parts(key, image_name, per_image, result_folder):
    if key == 'items':
        for image in per_image:
            if image["image_name"] == image_name:
                return image.get(key)

    if key == 'remote_url':
        all_masks = pathlib.Path(result_folder).parent / 'all_masks'
        for frame in all_masks.glob('*.jpg'):
            if frame.name == image_name:
                return frame
    return



from django.conf.urls import url
from django.views.decorators.csrf import csrf_exempt

from rebotics_sdk.hook import views


app_name = 'rebotics_sdk.hook'

urlpatterns = [
    url(r'^webhook/(?P<retailer_code>\w+)/$', csrf_exempt(views.WebhookHandlerViewSet.as_view()), name='webhook'),
]



import base64
import hmac
from collections import OrderedDict
from hashlib import sha256
from typing import TYPE_CHECKING

from rebotics_sdk.advanced import remote_loaders
from rebotics_sdk.providers import PageResult, ReboticsBaseProvider, remote_service

if TYPE_CHECKING:
    from requests.models import Response


class FVMProvider(ReboticsBaseProvider):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.codename = None
        self.secret_key = None

    def set_retailer_identifier(self, codename: str, secret_key: str) -> None:
        self.headers['x-retailer-codename'] = codename
        self.headers['x-retailer-secret-key'] = secret_key
        self.codename = codename
        self.secret_key = secret_key

    @remote_service('/api/synchronization/')
    def send_products_file_to_fvm(self, retailer_codename: str, products_count: int, file_url: str) -> OrderedDict:
        json = {
            'retailer_codename': retailer_codename,
            'products_count': products_count,
            'file': file_url
        }
        return self.session.post(data=json)

    @remote_service('/api/token-auth/', raw=True)
    def token_auth(self, username, password, verification_code=None):
        payload = dict(
            username=username,
            password=password,
        )
        if verification_code is not None:
            payload['verification_code'] = verification_code

        response = self.session.post(data=payload)
        self.headers['Authorization'] = 'Token %s' % response.json()['token']
        return response

    @remote_service('/api/files/virtual/')
    def create_virtual_upload(self, filename):
        return self.session.post(
            json={
                "filename": filename
            }
        )

    @remote_service('/api/files/{pk}/finish/')
    def finish(self, pk):
        return self.session.post(pk=pk)

    def upload_file(self, file_io, filename):
        file_io.seek(0)
        request = self.create_virtual_upload(filename)
        remote_loaders.upload(
            request['destination'],
            file_io
        )
        return self.finish(request['id'])

    @remote_service('/api/previews/{barcode}/')
    def get_preview_permalink(self, barcode):
        assert barcode.isdigit()
        url = self.build_url(self.session.url.format(barcode=barcode))
        if self.secret_key and self.codename:
            token = self.generate_token(url)
            return f'{url}?token={token}&codename={self.codename}'
        else:
            return url

    def generate_token(self, path):
        # You need this function to generate a token to use it, for example, in the authentication
        # You should call this function with an arg "path" for example: 'api/previews/'
        # With codename, path and secret_key, it will generate a token
        # Token will contain numbers, letters and symbols

        if self.secret_key is None or self.codename is None:
            raise AttributeError(
                "Identifiers for token generation are not present. Set secret_key and codename attributes or call .set_retailer_identifier method")
        lhmac = hmac.new(self.secret_key.encode('utf-8'), digestmod=sha256)
        string_to_sign = f'{self.codename}/{path}'
        lhmac.update(string_to_sign.encode('utf-8'))
        b64 = base64.urlsafe_b64encode(lhmac.digest())
        return b64

    @remote_service('/api/previews/')
    def upload_preview(self, barcode, image_url):
        return self.session.post(json={
            'barcode': barcode,
            'image_url': image_url
        })

    @remote_service('/api/products/')
    def get_products_list(self, page=1, page_size=100):
        params = {
            'page': page,
            'page_size': page_size,
        }
        return PageResult(self.session.get(params=params))

    @remote_service('/api/products/')
    def create_product(self, code, title):
        return self.session.post(json={
            'code': code,
            'title': title,
        })

    @remote_service('/api/products/{id}/')
    def update_product(self, id_, code, title):
        return self.session.put(
            id=id_,
            json={
                'code': code,
                'title': title,
            }
        )

    @remote_service('/api/rcdb/latest/')
    def get_latest_rcdb(self, model_codename, retailer_codename=None, with_images=False):
        payload = {
            'model_codename': model_codename,
            'with_images': with_images,
        }
        if retailer_codename is not None and self.codename is None:
            payload['retailer_codename'] = retailer_codename
        return self.session.post(json=payload)

    @remote_service('/api/rcdb/refresh/')
    def rcdb_refresh(self, model_codename, retailer_codename, with_images=False):
        # only admin can do that
        return self.session.post(json={
            'model_codename': model_codename,
            'retailer_codename': retailer_codename,
            'with_images': with_images,
        })

    @remote_service('/api/rcdb/{id}/')
    def get_rcdb_by_id(self, id_):
        return self.session.get(id=id_)

    @remote_service('/api/rcdb/')
    def save_rcdb(self, file_id, features_count, archive_size, model_codename, retailer_codename, with_images=False):
        return self.session.post(json={
            'file_id': file_id,
            'features_count': features_count,
            'archive_size': archive_size,
            'with_images': with_images,
            'model_codename': model_codename,
            'retailer_codename': retailer_codename,
        })

    @remote_service('/', raw=True, retries=5)
    def download(self, url, destination=None, progress_bar=False):
        if not url.startswith(self.host):
            url = self.build_url(url)

        def process_file_response(resp: 'Response', **request_arguments) -> None:
            downloaded_file = None
            if resp.ok:
                downloaded_file = remote_loaders.process_file_response(resp, destination, progress_bar)
            resp._downloaded_file = downloaded_file

        response = self.session.get(url=url, stream=True, hooks={'response': process_file_response})
        return response._downloaded_file  # noqa



from datetime import datetime

import pandas as pd
import numpy as np


PLANOGRAM_COLUMNS = [
    'planogram_number', 'planogram_title',
    'category_number', 'category_title',
    'bay_report_id',
    'aisle', 'section', 'shelf', 'shelf_position',
    'x', 'y',
    'upc', 'product_name',
]


def make_planogram_df(scan_data, planogram_number=None):
    if planogram_number is None:
        planogram_number = 'reverse_planogram_scan_{}'.format(scan_data['id'])

    products_map = {
        c["upc"]: c for c in scan_data['plu']
    }

    bay_map = {
        sb["id"]: sb for sb in scan_data['shelf_bays']
    }
    store = scan_data['store']

    flattened = []
    for item in scan_data['items']:
        product = products_map.get(item['upc'])
        flattened.append({
            'planogram_number': planogram_number,
            'planogram_title': planogram_number,
            'store_id': store.get('custom_id'),
            'category_number': scan_data.get('category_number'),
            'category_title': scan_data.get('category_title'),
            'bay_report_id': item['shelf_bay_id'],
            'aisle': bay_map[item['shelf_bay_id']]['aisle'],
            'section': bay_map[item['shelf_bay_id']]['section'],
            'shelf': item['display_shelf'],
            'shelf_position': item['position'],
            'x': item['x'],
            'y': item['y'],
            'upc': item['upc'],
            'product_name': product.get('title', "") if product else ""
        })

    df = pd.DataFrame(flattened, columns=PLANOGRAM_COLUMNS, index=None, dtype=np.dtype(object))
    pog_d = dict()

    for i, item in df.iterrows():
        key = "{aisle}|{section}|{shelf}|{x}|{upc}".format(
            aisle=item['aisle'],
            section=item['section'],
            shelf=item['shelf'],
            x=item['x'],
            upc=item['upc'],
        )
        d = pog_d.get(key)
        if d is None:
            item['facings_high'] = 1
            pog_d[key] = item
        else:
            d['facings_high'] = d.get('facings_high', 1) + 1

    df['facings_wide'] = 1
    df['facings_deep'] = 1
    df['facing_type'] = 'item'
    df['section_direction'] = 'left_to_right'
    df['shelf_direction'] = 'top_to_down'
    df['position_direction'] = 'left_to_right'
    df['from_date'] = datetime.now().strftime("%Y-%m-%d")
    df['is_zig_zag'] = 0
    return df



import pathlib

try:
    import click
except ImportError:
    raise Exception("To use authenticated role provider you have to install rebotics_sdk[shell]")

from rebotics_sdk.cli.utils import app_dir, ReboticsScriptsConfiguration

from rebotics_sdk.providers import \
    RetailerProvider, CvatProvider, AdminProvider, DatasetProvider, FVMProvider, HawkeyeProvider


PROVIDER_NAME_TO_CLASS = {
    'retailer': RetailerProvider,
    'cvat': CvatProvider,
    'admin': AdminProvider,
    'dataset': DatasetProvider,
    'fvm': FVMProvider,
    'hawkeye': HawkeyeProvider,
}


class AuthenticatedRoleProvider:
    def __init__(self):
        self.app_path = pathlib.Path(app_dir)
        assert self.app_path.exists(), "App path from cli should exist and be available"

    def get_provider(self, provider_name, role):
        config_provider = ReboticsScriptsConfiguration(
            self.app_path / f"{provider_name}.json",
            PROVIDER_NAME_TO_CLASS[provider_name]
        )
        return config_provider.get_provider(role)


def get_provider(provider_name, role):
    """
    get authenticated provider for given role.
    Intendent to be used in scripts provided by the rebotics team.

    :param provider_name: name of the provider. One of: retailer, cvat, admin, dataset, fvm, hawkeye
    :param role: role to use for authentication

    General usage:
    >>> from rebotics_sdk import get_provider
    >>> admin = get_provider('admin', 'r3dev')
    >>> admin.version()
    """
    return AuthenticatedRoleProvider().get_provider(provider_name, role)



import datetime


def test_api_heartbeat_invoke(runner, mocker_patch_provider):
    data = {"status": 200, "actions": []}
    mocker_patch_provider.save_camera_heartbeat.side_effect = lambda *args: data
    from rebotics_sdk.cli.hawkeye import api
    result = runner.invoke(
        api,
        [
            'heartbeat',
            '-c', 'test_camera',
            '-b', '0.55',
            '-w', '4.00',
            '-t', '2021-09-16T10:33:45.582912'
        ]
    )
    mocker_patch_provider.save_camera_heartbeat.assert_called_once_with(
        'test_camera',
        0.55,
        4.00,
        datetime.datetime(2021, 9, 16, 10, 33, 45, 582912).isoformat(),
    )
    assert result.exit_code == 0
    assert not result.exception
    assert str(data) in result.output.strip()


def test_api_heartbeat_invoke_without_camera(runner, mocker_patch_provider):
    from rebotics_sdk.cli.hawkeye import api
    result = runner.invoke(
        api,
        [
            'heartbeat',
            '-b', '0.55',
            '-w', '4.00',
            '-t', '2021-09-16T10:33:45.582912'
        ]
    )
    assert result.exit_code == 2
    assert result.exception
    assert "Missing option '-c' / '--camera'" in result.output.strip()


def test_api_fixture_invoke(runner, mocker_patch_provider):
    data = {
        "id": 1,
        "created": datetime.datetime(2021, 9, 16, 10, 33, 45, 582912),
        "modified": datetime.datetime(2021, 9, 16, 10, 33, 45, 582912),
        "store_id": "storetest",
        "aisle": "aisletest",
        "section": "sectiontest",
        "retailer": "testretailer"
    }
    mocker_patch_provider.save_fixture.side_effect = lambda *args: data
    from rebotics_sdk.cli.hawkeye import api
    result = runner.invoke(
        api,
        [
            'fixture',
            '-i', 'storetest',
            '-r', 'testretailer',
            '-a', 'aisletest',
            '-s', 'sectiontest'
        ]
    )
    mocker_patch_provider.save_fixture.assert_called_once_with(
        'storetest',
        'aisletest',
        'sectiontest',
        'testretailer'
    )
    assert result.exit_code == 0
    assert not result.exception
    assert str(data) in result.output


def test_api_fixture_invoke_without_option(runner, mocker_patch_provider):
    from rebotics_sdk.cli.hawkeye import api
    result = runner.invoke(
        api,
        [
            'fixture',
            '-i', 'storetest'
        ]
    )
    assert result.exit_code == 2
    assert result.exception
    assert "Missing option" in result.output


def test_get_fixtures_invoke(runner, mocker_patch_provider):
    data = []
    mocker_patch_provider.get_fixtures.side_effect = lambda *args: data
    from rebotics_sdk.cli.hawkeye import api
    result = runner.invoke(
        api,
        ['fixtures']
    )
    mocker_patch_provider.get_fixtures.assert_called()
    assert result.exit_code == 0
    assert not result.exception
    assert str(data) in result.output.strip()


def test_api_camera_invoke(runner, mocker_patch_provider):
    data = {
        "id": 1,
        "created": datetime.datetime(2021, 9, 16, 10, 33, 45, 582912),
        "modified": datetime.datetime(2021, 9, 16, 10, 33, 45, 582912),
        "camera_id": "testcameraid",
        "added_by": 1,
        "aisle": 1,
        "section": 1
    }
    mocker_patch_provider.create_shelf_camera.side_effect = lambda *args: data
    from rebotics_sdk.cli.hawkeye import api
    result = runner.invoke(
        api,
        [
            'camera',
            '-c', 'testcameraid',
            '-a', '1',
            '-f', '1'
        ]
    )
    mocker_patch_provider.create_shelf_camera.assert_called_once_with(
        'testcameraid',
        1,
        1
    )
    assert result.exit_code == 0
    assert not result.exception
    assert str(data) in result.output.strip()


def test_api_camera_invoke_without_option(runner, mocker_patch_provider):
    from rebotics_sdk.cli.hawkeye import api
    result = runner.invoke(
        api,
        [
            'camera',
            '-c', 'testcameraid'
        ]
    )
    assert result.exit_code == 2
    assert result.exception
    assert "Missing option" in result.output.strip()


def test_get_list_shelf_cameras(runner, mocker_patch_provider):
    mocker_patch_provider.get_shelf_cameras.side_effect = lambda *args: []
    from rebotics_sdk.cli.hawkeye import api
    result = runner.invoke(
        api,
        ['cameras']
    )
    mocker_patch_provider.get_shelf_cameras.assert_called()
    assert result.exit_code == 0
    assert not result.exception
    assert str([]) in result.output.strip()


def test_api_capture_invoke(runner, mocker_patch_provider):
    data = {"Message": "Capture created"}
    mocker_patch_provider.save_capture.side_effect = lambda *args: data
    from rebotics_sdk.cli.hawkeye import api
    result = runner.invoke(
        api,
        [
            'capture',
            '-c', 'test_camera_id',
            '-f', '/shelf_camera/testcameraid/2020/02/01/filename_11:01.jpeg',
            '-b', 'sftp-bucket'
        ]
    )
    mocker_patch_provider.save_capture.assert_called_once_with(
        'test_camera_id',
        '/shelf_camera/testcameraid/2020/02/01/filename_11:01.jpeg',
        'sftp-bucket'
    )
    assert result.exit_code == 0
    assert not result.exception
    assert str(data) in result.output.strip()




# coding=utf-8
import datetime
import json
import os
import pathlib
import time
from multiprocessing.pool import Pool
from time import sleep

import click
import humanize
from prettytable import PrettyTable

from rebotics_sdk.cli.common import shell, configure, roles, set_token
from .draw import draw_annotated_drawing
from .renderers import format_full_table
from .utils import ReboticsCLIContext, app_dir, pass_rebotics_context, read_saved_role, process_role, \
    downloads_with_threads, save_masks, get_segmentation, refresh_url, get_segmentation_mode
from ..advanced.flows import FileUploadError
from ..advanced.packers import ClassificationDatabasePacker, ClassificationDatabaseException, \
    DuplicateFeatureVectorsException
from ..providers import AdminProvider, RetailerProvider
from ..utils import uri_validator, get_filename_from_url

OPENCV_INSTALLED = True
try:
    import cv2
except ModuleNotFoundError:
    OPENCV_INSTALLED = False


@click.group()
@click.option('-f', '--format', default='table', type=click.Choice(['table', 'id', 'json']), help='Result rendering')
@click.option('-v', '--verbose', is_flag=True, help='Enables verbose mode')
@click.option('-c', '--config', type=click.Path(), default='admin.json', help="Specify what config.json to use")
@click.option('-r', '--role', default=lambda: read_saved_role('admin'), help="Key to specify what admin to use")
@click.version_option()
@click.pass_context
def api(ctx, format, verbose, config, role):
    """
    Admin CLI tool to communicate with dataset API
    """
    process_role(ctx, role, 'admin')
    ctx.obj = ReboticsCLIContext(
        'admin',
        role,
        format,
        verbose,
        os.path.join(app_dir, config),
        provider_class=AdminProvider
    )


def get_retailer_version_task(retailer_dict):
    retailer_provider = RetailerProvider(host=retailer_dict['host'], retries=1, timeout=5)
    try:
        response = retailer_provider.version()
        version = response['version']
        uptime = humanize.naturaldelta(datetime.timedelta(seconds=int(response['uptime'])))
    except Exception:
        version = 'not working'
        uptime = '---'

    d = [
        retailer_dict['codename'],
        retailer_dict['title'],
        version,
        uptime,
        retailer_dict['host'],
    ]
    return d


@api.command()
@click.argument('title', type=click.STRING)
@click.argument('group', type=click.STRING)
@click.argument('description', type=click.STRING)
@click.argument('yaml_template', type=click.File())
@pass_rebotics_context
def save_test_case(ctx, title, group, description, yaml_template):
    """
    Save test case in core.
    ARGS:
        title - STR,
        group - STR,
        description - STR,
        yaml_template - File
    """
    click.echo('Save core test case')
    yaml_template = yaml_template.read()

    try:
        ctx.provider.create_core_test_case(title=title,
                                           group=group,
                                           description=description,
                                           yaml_template=yaml_template)
        click.echo('Done')
    except Exception as exc:
        raise click.ClickException(str(exc))


@api.command()
@click.argument('title', type=click.STRING)
@click.argument('group', type=click.STRING)
@click.argument('description', type=click.STRING)
@click.argument('yaml_template', type=click.File())
@click.option('--target', type=click.Path(exists=False, file_okay=True, dir_okay=True), default=None)
@pass_rebotics_context
def retrieve_test_case_data(ctx, title, group, description, yaml_template, target):
    """
    Retrieve data for test case from core.
    ARGS:
        title - STR,
        group - STR,
        description - STR,
        yaml_template - File,
        target - Path
    """
    click.echo('Rendering test_case data')
    yaml_template = yaml_template.read()

    try:
        res = ctx.provider.retrieve_core_test_case_data(title=title,
                                                        group=group,
                                                        description=description,
                                                        yaml_template=yaml_template)

        if target is None:
            ctx.format_result(res)
        else:
            target = pathlib.Path(target)
            if target.is_dir():
                target = target / f"test_case_{time.strftime('%d-%m-%Y_%H:%M')}.json"
            with open(target, 'w') as fd:
                fd.write(json.dumps(res, indent=2))
                click.echo(f'Download to {target}')
    except Exception as exc:
        raise click.ClickException(str(exc))


@api.command()
@click.argument('retailer')
@pass_rebotics_context
def configurations(ctx, retailer):
    """
        Fetch retailer configurations variables
    """
    try:
        res = ctx.provider.get_configurations(retailer)
        ctx.format_result(res)
    except Exception as exc:
        raise click.ClickException(str(exc))


@api.command()
@click.argument('retailer')
@pass_rebotics_context
def mobile_configurations(ctx, retailer):
    """
    Fetch configurations for mobile application for retailer
    """
    res = ctx.provider.get_mobile_configurations(retailer)
    ctx.format_result(res)


@api.command()
@click.option('-n', '--notify', is_flag=True)
@click.option('-d', '--delay', type=click.INT, default=60)
@pass_rebotics_context
def retailer_versions(ctx, notify, delay):
    """Fetch retailer versions and their meta information"""
    click.echo("For better experience use: \n"
               "https://versions.fyn.rocks/ and https://tracker.fyn.rocks/")

    if notify:
        if ctx.verbose:
            click.echo('Using notify option', err=True)

        try:
            from pynotifier import Notification
            Notification(
                title='Subscribed to the notifications',
                description='You will receive notifications for retailer updates',
            ).send()
        except ImportError:
            raise click.ClickException("You can't use notify function")

    provider = ctx.provider
    if ctx.verbose:
        click.echo('Fetching info from rebotics admin.', err=True)
    retailers = provider.get_retailer_list()
    prev_results = []
    results = []
    pool = Pool(len(retailers))

    while True:
        try:
            if ctx.verbose:
                click.echo('Fetching the retailer versions', err=True)
            results = pool.map(get_retailer_version_task, retailers)

            if not notify:
                break

            for prev_result in prev_results:
                retailer_codename = prev_result[0]
                previous_version = prev_result[2]
                for result in results:
                    if result[0] == retailer_codename:
                        current_version = result[2]
                        if previous_version != current_version:
                            notification_message = 'Retailer {} updated from version {} to {}'.format(
                                retailer_codename,
                                previous_version,
                                current_version
                            )
                            click.echo(notification_message)
                            Notification(
                                title=notification_message,
                                description='Current uptime is: {}'.format(result[3]),
                                duration=30,
                                urgency=Notification.URGENCY_CRITICAL,
                            ).send()
            del prev_results
            prev_results = results
            sleep(delay)
        except KeyboardInterrupt:
            break

    table = PrettyTable()
    table.field_names = ['codename', 'title', 'version', 'uptime', 'host']
    for result in results:
        table.add_row(result)
    click.echo(table)


@api.command()
@click.argument('retailer')
@click.option('--v2', is_flag=True, show_default=True, default=False, help="Use models v2 API")
@pass_rebotics_context
def models(ctx, retailer, v2):
    """
        Fetch and display retailer NN configurations
    """
    try:
        if v2:
            ctx.format_result(
                ctx.provider.get_models_conf_v2(codename=retailer)
            )
        else:
            ctx.format_result(
                ctx.provider.get_retailer_tf_models(codename=retailer)
            )
    except Exception as exc:
        raise click.ClickException(str(exc))


@api.command()
@click.argument('retailer', type=click.STRING)
@click.argument('url', type=click.STRING)
@pass_rebotics_context
def set_retailer_url(ctx, retailer, url):
    try:
        ctx.provider.update_host(retailer, url)
    except Exception as exc:
        raise click.ClickException(str(exc))
    else:
        click.echo('Set new host for retailer %s' % retailer)


@api.command()
@click.option('-r', '--retailer', help="Retailer codename", prompt=True)
@click.option('-m', '--model', help="Model codename", prompt=True)
@click.option('-i', '--images', type=click.Path(exists=True, file_okay=False, dir_okay=True))
@click.option('-l', '--labels', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.option('-f', '--features', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.option('-c', '--check-duplicates', is_flag=True, help="MD5 checker simple")
@click.option('-m', '--model-type', type=str, default='facenet')
@click.option('-n', '--model-codename', type=str)
@pass_rebotics_context
def pack_and_import_classification_db(
    ctx,
    retailer,
    model,
    images,
    labels,
    features,
    check_duplicates,
    model_type,
    model_codename,
):
    """Pack classification database to single .rcdb file and import it to the Rebotics Admin

\b
db/
├── custom_folder
│   ├── image_2.png
│   └── image_1.png
├── features.txt
└── labels.txt

    It is a single step command with equivalent of running two commands:

admin pack-classification-db --features features.txt --labels labels.txt --images ./custom_folder/ --target classification.rcdb

admin import_classification_db --retailer delta --model test_code classification.rcdb
    """
    ext = ClassificationDatabasePacker.extension
    with ctx.provider.rcdb.import_flow(retailer, model, ext) as flow:
        if ctx.verbose:
            click.echo("Creating import request: \n"
                       "retailer: {}\n"
                       "model: {}\n"
                       "extension: {}".format(retailer, model, ClassificationDatabasePacker))

        packer = ClassificationDatabasePacker(
            destination=None,
            progress_bar=True,
            check_duplicates=check_duplicates,
            model_type=model_type,
            model_codename=model_codename,
        )
        if ctx.verbose:
            click.echo("Packing from provided values: \n"
                       "labels: {labels} \n"
                       "features: {features} \n"
                       "images: {images}".format(labels=labels, features=features, images=images))

        try:
            packed = packer.pack(labels, features, images)
            packed.seek(0)
        except DuplicateFeatureVectorsException as exc:
            handle_duplicates(ctx, exc.duplicates)
            raise click.ClickException(str(exc))
        except ClassificationDatabaseException as exc:
            raise click.ClickException(str(exc))

        if ctx.verbose:
            click.echo("Import request: {}".format(flow.import_request))

        try:
            flow.upload_file(packed, progress_bar=True)
        except FileUploadError as exc:
            click.echo(str(exc))
            raise

        click.echo("Import completed. Reference database id: {}".format(flow.import_request['id']))


@api.command()
@click.option('-r', '--retailer', help="Retailer codename", prompt=True)
@click.option('-m', '--model', help="Model codename", prompt=True)
@click.argument('filepath', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@pass_rebotics_context
def import_classification_db(ctx, retailer, model, filepath):
    """Import rcdb file to Rebotics Admin. Example usage:

    admin import_classification_db --retailer delta --model test_code classification.rcdb

    """
    extension = os.path.split(filepath)[-1].split('.')[-1]
    if ctx.verbose:
        click.echo("Creating import request: \n"
                   "retailer: {}\n"
                   "model: {}\n"
                   "extension: {}".format(retailer, model, extension))

    with ctx.provider.rcdb.import_flow(retailer, model, extension) as flow:
        if ctx.verbose:
            click.echo("Import request: {}".format(flow.import_request))
        with open(filepath, 'rb') as packed:
            flow.upload_file(packed, progress_bar=True)

    click.echo("Import completed. Reference database id: {}".format(flow.import_request['id']))


def handle_duplicates(ctx, duplicates):
    unpacked = [
        [
            {
                'id': entry.index,
                'label': entry.label,
                'filename': entry.filename
            }
            for entry in group
        ] for group in duplicates
    ]
    if ctx.format == 'json':
        click.echo(json.dumps(unpacked, indent=2))
    elif ctx.format == 'table':
        for group in unpacked:
            format_full_table(group)
    elif ctx.format == 'id':
        click.echo("Duplicate id. New line separated for group; space separated inside of the group:")
        for group in duplicates:
            click.echo(" ".join([str(entry.index) for entry in group]))
    else:
        for group in duplicates:
            for entry in group:
                click.echo("{entry.index} {entry.label} {entry.filename}".format(entry=entry))
            click.echo('=' * 100)


@api.command()
@click.option('-i', '--images', type=click.Path(exists=True, file_okay=False, dir_okay=True))
@click.option('-l', '--labels', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.option('-f', '--features', type=click.Path(exists=True, file_okay=True, dir_okay=False))
@click.option('-c', '--check-duplicates', is_flag=True, help="MD5 checker simple")
@click.option('-m', '--model-type', type=str, default='facenet')
@click.option('-n', '--model-codename', type=str)
@click.argument('target', type=click.Path(exists=False, file_okay=True, dir_okay=True), default='.')
@pass_rebotics_context
def pack_classification_db(ctx, images, labels, features, target, check_duplicates, model_type, model_codename):
    """Pack classification database to single .rcdb file

\b
db/
├── custom_folder
│   ├── image_2.png
│   └── image_1.png
├── features.txt
└── labels.txt

    Example usage:
        admin pack-classification-db --features features.txt --labels labels.txt --images ./custom_folder/ --target classification.rcdb

    """
    target = pathlib.Path(target)
    if target.is_dir():
        target = target / "classification_db{}.{}".format(
            datetime.datetime.now().strftime('%Y-%m-%dZ%H%M'),
            ClassificationDatabasePacker.extension
        )
    packer = ClassificationDatabasePacker(
        destination=target,
        progress_bar=True,
        check_duplicates=check_duplicates,
        model_type=model_type,
        model_codename=model_codename,
    )
    if ctx.verbose:
        click.echo("Packing from provided values: \n"
                   "labels: {labels} \n"
                   "features: {features} \n"
                   "images: {images}".format(labels=labels, features=features, images=images))
    try:
        packed = packer.pack(labels, features, images)
    except DuplicateFeatureVectorsException as exc:
        handle_duplicates(ctx, exc.duplicates)
        raise click.ClickException(str(exc))
    except ClassificationDatabaseException as exc:
        raise click.ClickException(str(exc))
    click.echo('Written to {}'.format(packed))


@api.group(name='rcdb')
def rcdb_group():
    click.echo("RCDB in rebotics admin is deprecated and will be removed in version 0.13.0")


@rcdb_group.command(name="previews-fs")
@click.argument("argument", type=click.STRING)
@click.option('-o', "--output", type=click.Path(exists=True, dir_okay=True, file_okay=False), default=".")
@click.option('-c', '--concurrency', default=8, type=click.INT)
@pass_rebotics_context
def rcdb_previews_fs(ctx, argument, output, concurrency):
    """
    Prepare previews file structure by the id, filepath or file url of the JSON formatted backup
    Make sure that extension is a valid JSON and extension in the admin is set like that
    """
    click.echo("Use `fvm rcdb unpack` command instead")


@api.group()
def stitching():
    pass


@stitching.command(name='setup')
@click.argument("argument", type=click.STRING)
@click.option('-o', "--output", type=click.Path(dir_okay=True, file_okay=False))
@click.option('-c', '--concurrency', default=8, type=click.INT)
@click.option('-d', '--draw', is_flag=True, help='To draw annotated frames')
@pass_rebotics_context
def stitching_setup(ctx, argument, concurrency, output, draw):
    """
    argument - id of stitching debug data from admin, or json file
    """
    if draw and not OPENCV_INSTALLED:
        raise click.ClickException(' OpenCv not found, but can be installed with: pip install opencv-python ')

    if output:
        output = pathlib.Path(output).absolute()
    else:
        output = pathlib.Path('.').absolute()

    output = output / f"stitching_{argument.replace(' ', '_')}"
    output.mkdir(parents=True, exist_ok=True)
    ctx.verbose_log(f"Target directory is: {output}")
    filepath = output / "definition.json"

    if argument.isdigit():
        backup_id = int(argument)
        callback_data = ctx.provider.get_core_callback_data(backup_id)
        definition = callback_data['data']

        ctx.verbose_log("Saving definition locally")
        with open(filepath, 'w') as fp:
            json.dump(definition, fp)

    elif uri_validator(argument):
        # argument is a url
        raise click.ClickException("not supported")
    else:
        ctx.verbose_log("Assuming we have a filepath")
        filepath = argument
        try:
            with open(filepath, 'rb') as fp:
                definition = json.load(fp)
        except Exception:
            raise click.ClickException(f"File is not supported. Please check it at {filepath}")

    if 'data' in definition:
        definition = definition['data']

    if ctx.verbose:
        ctx.format_result(definition['context'], force_format='json')

    files_to_load = []  # pair of remote_url to local path

    segmentation_filepath = output / "image_segmentation_debug.json"
    segmentation_url = definition.get("image_segmentation_debug", None)
    if segmentation_url is not None:
        files_to_load.append([segmentation_url, segmentation_filepath])
    else:
        draw = False
        ctx.verbose_log("Cannot draw bboxes, no image segmentation debug file is available")

    meta_df_filepath = output / "task_meta_df.csv"
    meta_df_url = definition.get("task_meta_df", None)
    if meta_df_url is not None:
        files_to_load.append([meta_df_url, meta_df_filepath])

    for call_count, call_data in enumerate(definition['calls']):
        call_folder = output / f"call_{call_count}"
        call_folder.mkdir(parents=True, exist_ok=True)

        input_data = call_data['input']

        stitching_input = {
            **input_data,
            'frame_paths': [],
            'output_file': str(call_folder),
            'result_image': str(call_folder / "local_stitching.jpeg"),
        }

        del stitching_input['frame_urls']
        # for backward compatible
        if 'annotated_frames' in stitching_input:
            del stitching_input['annotated_frames']

        # add stitching result image
        try:
            files_to_load.append([
                call_data['image_url'], call_folder / "server_stitched_image.jpeg"
            ])
        except Exception:
            ctx.verbose_log("No stitching image is available")

        if input_data.get('mask_od_url'):
            mask_od_path = call_folder / "mask_od.json"
            stitching_input['mask_od'] = str(mask_od_path)
            files_to_load.append([
                input_data['mask_od_url'], mask_od_path
            ])

        # register input files
        frames_folder = call_folder / "frames"
        frames_folder.mkdir(parents=True, exist_ok=True)

        for frame_url in input_data['frame_urls']:
            frame_path = frames_folder / get_filename_from_url(frame_url)
            stitching_input['frame_paths'].append(str(frame_path))
            files_to_load.append([
                frame_url, frame_path
            ])

        frames_folder = call_folder / "annotated_frames"
        frames_folder.mkdir(parents=True, exist_ok=True)

        annotated_frames = input_data.get('annotated_frames', [])
        if annotated_frames:
            for frame_url in annotated_frames:
                frame_path = frames_folder / get_filename_from_url(frame_url)
                files_to_load.append([
                    frame_url, frame_path
                ])

        ctx.verbose_log(f'Creating a stitching input for run #{call_count} into {call_folder}')
        with open(str(call_folder / "stitching_input.json"), 'w') as fp:
            json.dump(stitching_input, fp)

    ctx.verbose_log(f"Registered files for downloading: {len(files_to_load)}")

    downloads_with_threads(files_to_load, concurrency=len(files_to_load))
    click.echo(f"Stitching setup complete in: {output}")

    if draw:
        segmentation_per_image = get_segmentation(segmentation_filepath)
        mode = get_segmentation_mode(segmentation_per_image)
        if mode == 'remote_url':
            refreshed_urls = refresh_url(ctx.provider.refresh_url, segmentation_per_image)
            # save masks in output/all_masks
            save_masks(output, refreshed_urls)

        draw_annotated_drawing(
            root_folder=output,
            segmentation_per_image=segmentation_per_image,
            mode=mode,
            concurrency=concurrency
        )


api.add_command(shell, 'shell')
api.add_command(roles, 'roles')
api.add_command(configure, 'configure')
api.add_command(set_token, 'set_token')



import logging

from .base import ReboticsBaseProvider, remote_service

logger = logging.getLogger(__name__)


class CvatProvider(ReboticsBaseProvider):
    """
    See <cvat-url>/api/docs/#tag/retailer_import
    for more detailed examples.
    """

    @remote_service('/api/retailer_import')
    def start_retailer_import(self, data):
        """
        Data example:
        {
          "image_quality": 80,  # 0 - 100
          "segment_size": 20,    # >= 0
          "workspace": "RetechLabs",
          "export_by": "username",
          "retailer_codename": "delta",
          "images": [
            {
              "items": [
                {
                  "lowerx": 0,
                  "lowery": 0,
                  "upperx": 0,
                  "uppery": 0,
                  "label": "string",
                  "points": "string",
                  "type": "string",
                  "upc": "string"
                }
              ],
              "image": "http://example.com",
              "planogram_title": "string",
              "processing_action_id": 0,
              "price_tags": [
                {
                  "lowerx": 0,
                  "lowery": 0,
                  "upperx": 0,
                  "uppery": 0,
                  "label": "string",
                  "points": "string",
                  "type": "string",
                  "upc": "string"
                }
              ]
            }
          ]
        }

        Response example:
        {
          "task_id": 0,
          "preview": None,
          "images": None,
          "status": None
        }
        """
        return self.session.post(json=data)

    @remote_service('/api/retailer_import/{id}')
    def check_import_progress(self, task_id):
        """
        Does not accept any data.

        Response example:
        {
          "task_id": 0,
          "preview": "http://example.com",
          "images": [
            {
              "id": 0,
              "image": "http://example.com"
            }
          ],
          "status": {
            "state": "Queued",
            "message": "",
            "progress": 0
          }
        }
        """
        return self.session.get(id=task_id)



from io import BytesIO
from unittest import TestCase
from unittest.mock import patch

import requests_mock

from rebotics_sdk.providers import RetailerProvider, ProviderHTTPServiceException


@patch('time.sleep', new=lambda *args, **kwargs: None)
class ProcessingUpload(TestCase):
    def setUp(self) -> None:
        self.file_io = BytesIO()
        self.provider = RetailerProvider(host='http://test.com')

    @patch('rebotics_sdk.providers.base.ProviderRequestProxy.format_return_data')
    def test_processing_upload(self, mock_func):
        with requests_mock.Mocker() as m:
            m.post('http://test.com/api/v4/processing/upload/')
            with patch.object(self.file_io, 'seek') as mock_file:
                self.provider.processing_upload(store_id=1, input_file=self.file_io)
                self.assertEqual(2, mock_file.call_count)

    @patch('rebotics_sdk.providers.base.ProviderRequestProxy.format_return_data')
    def test_processing_upload_with_empty_files(self, mock_func):
        with requests_mock.Mocker() as m:
            m.post('http://test.com/api/v4/processing/upload/', status_code=502)
            with patch.object(self.file_io, 'seek') as mock_file:
                with self.assertRaises(ProviderHTTPServiceException) as error:
                    self.provider.processing_upload(store_id=1, input_file=self.file_io)
            self.assertEqual(7, mock_file.call_count)
            self.assertTrue('Failed to do request after ' in error.exception.args[0])



from django.conf import settings

REBOTICS_RETAILER_SERVICE_HOST = getattr(settings, 'REBOTICS_RETAILER_SERVICE_HOST', '')
REBOTICS_RETAILER_SERVICE_TOKEN = getattr(settings, 'REBOTICS_RETAILER_SERVICE_TOKEN', '')



from functools import partial
from multiprocessing import Pool
from pathlib import Path

from six.moves.urllib.parse import urlparse

from rebotics_sdk.advanced import remote_loaders
from rebotics_sdk.constants import RCDB


class FileUploadError(Exception):
    def __init__(self, msg, response):
        super(FileUploadError, self).__init__(msg)
        self.response = response


class RCDBImportFlow(object):
    def __init__(self, rcdb_interface, retailer, retailer_model, extension):
        """
        Process wrapper with automatic API hooks

        :param ReboticsClassificationDatabase rcdb_interface:
        :param str retailer:
        :param str retailer_model:
        :param str extension:
        """
        self.rcdb = rcdb_interface
        self.import_request = self.rcdb.data_import(retailer, retailer_model, extension)
        self.rcdb_id = self.import_request['id']

    def __enter__(self):
        self.rcdb.update(self.rcdb_id, status=RCDB.EXPORT_IN_PROGRESS)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type or exc_val:
            self.rcdb.update(self.rcdb_id, status=RCDB.EXPORT_ERROR)
        else:
            self.rcdb.update(self.rcdb_id, status=RCDB.EXPORT_DONE)

    def upload_file(self, packed, progress_bar=False):
        destination = self.import_request['destination']
        response = remote_loaders.upload(destination, packed, progress_bar=progress_bar)
        if response.status_code > 300:
            raise FileUploadError(
                'Failed to upload destination. Response: {}. {}'.format(response.status_code, response.content),
                response)


class ReboticsClassificationDatabase(object):
    def __init__(self, *args, **kwargs):
        self.provider = kwargs.get('provider', None)

    def create(self, retailer, model, backup_type='feature_vectors', extension='rcdb'):
        return self.provider.rcdb_create({
            'retailer': retailer,
            'facenet_model': model,
            'extension': extension,
            'backup_type': backup_type,
        })

    def data_import(self, retailer, model, extension='zip', **kwargs):
        return self.provider.create_classification_database_import(retailer, model, extension, **kwargs)

    def update(self, id_, status=RCDB.EXPORT_DONE, entries_count=0):
        assert status in (c[0] for c in RCDB.EXPORT_STATUS_CHOICES)
        return self.provider.rcdb_update(id_, {
            'status': status,
            'entries_count': entries_count
        })

    def get(self, id_):
        return self.provider.rcdb_get(id_)

    def import_flow(self, retailer, retailer_model, extension):
        return RCDBImportFlow(
            self, retailer, retailer_model, extension
        )


class FeatureVectorFlow(object):
    def __init__(self, provider, session):
        self.provider = provider
        self.session = session

    def get(self):
        return self.session.get()

    def export(self, triggered_by='sdk', source_model='', result_model='previews_backup', batch_size=50000):
        d = dict(
            triggered_by=triggered_by,
            source_model=source_model,
            result_model=result_model,
            batch_size=batch_size
        )
        return self.session.post(json=d)



import dataclasses
import io
import json
import logging
import os
import pathlib
import re
import sys
import typing
import uuid
import zipfile
from collections import namedtuple, defaultdict
from datetime import datetime
from hashlib import md5
from io import BytesIO, StringIO
from typing import Union, AnyStr

import py7zr

import tqdm

import rebotics_sdk
from rebotics_sdk.utils import validate_files, read_lines
from rebotics_sdk.constants import RCDBBaseFileNames

logger = logging.getLogger(__name__)

ClassificationEntry = namedtuple('ClassificationEntry', ['label', 'feature', 'image', 'filename', 'index'])
ImageEntry = namedtuple('ImageEntry', ['filename', 'filepath', 'order'])


class VirtualClassificationEntry(object):
    def __init__(self, label, feature, image_url, index=None):
        self.label = label
        self.feature = feature
        self.image_url = image_url
        self.index = index

    def to_dict(self):
        return {
            'label': self.label,
            'feature': self.feature,
            'image_url': self.image_url,
            'index': self.index
        }

    @classmethod
    def from_dict(cls, d):
        return cls(
            label=d['label'],
            feature=d['feature'],
            image_url=d['image_url'],
            index=d['index']
        )


NUMERIC_VALUE = re.compile(r'\d+')


class ClassificationDatabaseException(Exception):
    pass


class DuplicateFeatureVectorsException(ClassificationDatabaseException):
    def __init__(self, msg, duplicates):
        super(DuplicateFeatureVectorsException, self).__init__(msg)
        self.duplicates = duplicates


class UnpackException(ClassificationDatabaseException):
    def __init__(self, archive_name: AnyStr, data_keys: typing.List):
        msg = f'Archive {archive_name} has incorrect structure. Expected to have \'{RCDBBaseFileNames.LABELS}\', ' \
              f'\'{RCDBBaseFileNames.FEATURES}\' and \'{RCDBBaseFileNames.META}\' files among keys. ' \
              f'Provided keys are: {str(data_keys)}.'
        super(UnpackException, self).__init__(msg)
        self.archive_name = archive_name
        self.data_keys = data_keys


def extract_numeric(filename):
    found = NUMERIC_VALUE.findall(filename)
    if len(found) == 0:
        raise ClassificationDatabaseException('Images name should contain numeric value that represents the order. '
                                              'Instead of {} name was used'.format(filename))
    return int(found[0])


class BaseClassificationDatabasePacker:
    """
    Base class for classification database packing process.
    """
    extension = 'rcdb'

    def __init__(self, destination, **kwargs):
        self.meta_data = {
            'packed': datetime.now().strftime('%c'),
            'model_type': kwargs.get('model_type', 'facenet'),
            'model_codename': kwargs.get('model_codename'),
            'sdk_version': rebotics_sdk.__version__,
            'core_version': kwargs.get('core_version'),
            'fvm_version': kwargs.get('fvm_version')
        }

        if hasattr(destination, 'write') or hasattr(destination, 'read'):
            self.destination = destination  # file like object
        elif destination is None:
            self.destination = BytesIO()
        elif isinstance(destination, str) or isinstance(destination, pathlib.PurePath):
            # in python 2 it doesn't accept pathlib PurePath, but only os.path or str
            self.destination = str(destination)
            if not self.destination.endswith(self.extension):
                self.destination = "{}.{}".format(self.destination, self.extension)

    def pack(self, *args, **kwargs):
        """
        Function to pack data from arguments into an archive. Should accept set of parameters that describe input data,
        and write it all into archive that was set up in __init__.

        :returns: path to archive
        """
        raise NotImplementedError()

    def unpack(self):
        """
        Function to unpack existing archive. Unpacks the archive, then merges together entries from label and

        :returns: Returns a single classification entry iteratively (yield return)
        """
        raise NotImplementedError()

    def archive_file(self, **kwargs):
        """
        Function to return fully-initialized file-like object that provides interface to archive file.

        :returns: file-like object that represents the archive
        """
        raise NotImplementedError()

    def read_meta(self):
        """
        Function to read metadata from sourcing archive, if it exists.
        """
        raise NotImplementedError()


class BaseZipClassificationDatabasePacker(BaseClassificationDatabasePacker):
    version = 0
    extension = 'zip'

    def __init__(
        self,
        source=None,
        destination=None,
        progress_bar=False,
        check_duplicates=False,
        compresslevel=1,
        *args,
        **kwargs
    ):
        if source is not None and destination is not None:
            raise ValueError("You can't sent source and destination at a same time.")

        super().__init__(destination, **kwargs)
        self.__compresslevel = compresslevel
        self.meta_data['packer_version'] = self.version

        self.source = source

        self.zip_io = None

        if self.source is not None:
            self.read_meta()

        self.check_duplicates = check_duplicates
        self.images = []
        self.progress_bar = progress_bar

    @property
    def compresslevel(self):
        return self.__compresslevel

    def read_lines(self, lines):
        return [x for x in lines.split('\n') if x]

    def zipfile(self, file, **kwargs):
        # if python version is
        params = dict(
            compression=zipfile.ZIP_DEFLATED,
            compresslevel=self.__compresslevel,
            allowZip64=True,
        )
        #
        # if (sys.version_info.major, sys.version_info.minor) >= (3, 7):
        #     # using the most aggressive compression
        #     params['compresslevel'] = 9  # default is 6

        # Compatibility with py36
        if sys.version_info < (3, 7):
            params.pop('compresslevel', None)

        params.update(kwargs)

        for compression_option in [zipfile.ZIP_STORED, zipfile.ZIP_DEFLATED]:
            try:
                params['compression'] = compression_option
                return zipfile.ZipFile(file, **params)
            except RuntimeError:
                pass

    archive_file = zipfile

    def __enter__(self):
        if self.zip_io is not None:
            return self

        # check for file to be not closed

        if self.source is not None:
            self.zip_io = self.zipfile(self.source, mode='r')
        elif self.destination is not None:
            self.zip_io = self.zipfile(self.destination, mode='w')
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.zip_io.close()
        self.zip_io = None

    def read_meta(self):
        with self as packer:
            self.meta_data = json.load(packer.zip_io.open('meta.json'))


class ZipDatabasePacker(BaseZipClassificationDatabasePacker):
    """
    Unified file format for packing classification database into single folder
    File format is the zip based archive with following :

    labels.txt
    features.txt
    meta.json
    """
    version = 0
    extension = 'zip'

    def pack(self, labels, features, *args, **kwargs):
        labels_io = StringIO("\n".join(labels))
        features_io = StringIO("\n".join(features))
        self.meta_data['count'] = len(labels)

        with self.zipfile(self.destination, mode='w') as zip_io:
            zip_io.writestr('labels.txt', labels_io.getvalue())
            zip_io.writestr('features.txt', features_io.getvalue())
            zip_io.writestr('meta.json', json.dumps(self.meta_data))
        return self.destination

    def unpack(self):
        if self.source is None:
            raise ClassificationDatabaseException("Can not unpack with empty source file")

        with self.zipfile(self.source, mode='r') as zip_io:
            try:
                self.meta_data = json.load(zip_io.open('meta.json'))
            except KeyError:
                # meta.json is not presented, working in the compatibility mode
                pass

            labels = self.read_lines(zip_io.read('labels.txt').decode('utf-8'))
            features = self.read_lines(zip_io.read('features.txt').decode('utf-8'))

            for index, (label, feature) in enumerate(zip(labels, features)):
                yield ClassificationEntry(
                    label, feature, None, None, index
                )


class ClassificationDatabasePacker(BaseZipClassificationDatabasePacker):
    """
    Unified file format for packing classification database into single folder
    File format is the zip based archive with following :

    labels.txt
    features.txt
    meta.json
    images/  - folder with images, this folder can be empty
    """

    version = 1
    extension = 'rcdb'
    images_extensions = {'jpeg', 'png', 'jpg'}

    def extract_meta_data(self, labels_path, features_path, images_folder):
        known_order = set()
        for f in pathlib.Path(images_folder).iterdir():
            if f.is_file():
                extension = f.suffix
                if extension in self.images_extensions:
                    raise ClassificationDatabaseException("Extension is not supported by packer: {}".format(f.name))
                numeric_order = extract_numeric(f.name)
                if numeric_order in known_order:
                    raise ClassificationDatabaseException("Numeric order collision. {}".format(f.name))

                known_order.add(numeric_order)
                self.images.append(ImageEntry(f.name, str(f), numeric_order))

        # repack based on the filename and it's numeric value
        self.images.sort(key=lambda i: i.order)  # sort by numeric value of the filename

        with open(labels_path, 'r') as labels_io, open(features_path, 'r') as features_io:
            labels = self.read_lines(labels_io.read())
            features = self.read_lines(features_io.read())

        labels_count = len(labels)
        features_count = len(features)
        files_count = len(self.images)

        if labels_count != features_count:
            raise ClassificationDatabaseException('Inconsistent count of labels and features. {}/{}'.format(
                labels_count, features_count
            ))

        if labels_count != files_count:
            raise ClassificationDatabaseException(
                'Inconsistent count of labels and features and files. {}/{}/{}'.format(
                    labels_count, features_count, files_count
                ))

        self.meta_data.update({
            'count': features_count,
            'images': [str(img.filename) for img in self.images]
        })

        if not self.check_duplicates:
            return

        # checking for the uniqueness of the FV
        features_map = defaultdict(list)
        for i, fv in tqdm.tqdm(enumerate(features), total=features_count, disable=not self.progress_bar, leave=False):
            md_ = md5(fv.encode('utf-8'))
            features_map[md_.hexdigest()].append(i)
        duplicates = []
        for same_fv_ids in features_map.values():
            if len(same_fv_ids) > 1:
                duplicates.append([
                    ClassificationEntry(
                        labels[id_],
                        features[id_],
                        None,
                        self.images[id_].filename,
                        id_
                    ) for id_ in same_fv_ids
                ])

        if duplicates:
            raise DuplicateFeatureVectorsException(
                "Detected {} duplicate groups.".format(len(duplicates)),
                duplicates=duplicates
            )

    def pack(self, labels, features, images, *args, **kwargs):
        self.extract_meta_data(labels, features, images)

        with self.zipfile(self.destination, mode='w') as zip_io:
            zip_io.write(labels, 'labels.txt')
            zip_io.write(features, 'features.txt')
            zip_io.writestr('meta.json', json.dumps(self.meta_data))

            for img in tqdm.tqdm(
                self.images,
                leave=False,
                disable=not self.progress_bar
            ):
                zip_io.write(img.filepath, str(pathlib.Path('images') / img.filename))

        return self.destination

    def unpack(self):
        """

        :return: generator of label, feature, image_io
        :rtype:self.meta_data
        """
        if self.source is None:
            raise ClassificationDatabaseException("Can not unpack with empty source file")

        with self.zipfile(self.source, mode='r') as zip_io:
            labels = self.read_lines(zip_io.read('labels.txt').decode('utf-8'))
            features = self.read_lines(zip_io.read('features.txt').decode('utf-8'))

            for index, (label, feature, image_name) in enumerate(zip(
                labels,
                features,
                self.meta_data['images']
            )):
                image = zip_io.read('images/{}'.format(image_name))
                yield ClassificationEntry(
                    label, feature, image, image_name, index
                )


@dataclasses.dataclass
class DataEntry:
    label: str
    feature: typing.List[float]
    uuid: typing.Optional[uuid.UUID]
    image_url: str


class VirtualClassificationDatabasePacker(BaseZipClassificationDatabasePacker):
    version = 2
    extension = 'rcdb'

    def __init__(self, *args, **kwargs):
        self.concurrency = kwargs.get('concurrency', None)
        if self.concurrency is None or self.concurrency < 1:
            self.concurrency = os.cpu_count()
        super(VirtualClassificationDatabasePacker, self).__init__(*args, **kwargs)

    def pack(self, labels, features, uuids, images, *args, **kwargs):
        with open(labels, 'r') as labels_io:
            self.meta_data['count'] = sum(1 for _ in labels_io)

        with self.zipfile(self.destination, mode='w') as zip_io:
            # writing those as a file
            zip_io.write(labels, 'labels.txt')
            zip_io.write(features, 'features.txt')
            zip_io.write(uuids, 'uuid.txt')
            zip_io.write(images, 'images.txt')

            zip_io.writestr('meta.json', json.dumps(self.meta_data))
        return self.destination

    def unpack(self) -> typing.Iterable[DataEntry]:
        for batch in self._data_entry_generator():
            yield from batch

    def get_features_count(self):
        with self.zipfile(self.source, mode='r') as zip_io:
            self.meta_data = json.load(zip_io.open('meta.json'))
        return self.meta_data.get('count')

    def _data_entry_generator(self):
        batch = []  # initial batch
        with self.zipfile(self.source, mode='r') as zip_io:
            labels_io = io.TextIOWrapper(zip_io.open("labels.txt"), encoding='utf-8')
            features_io = io.TextIOWrapper(zip_io.open("features.txt"), encoding='utf-8')
            uuids_io = io.TextIOWrapper(zip_io.open("uuid.txt"), encoding='utf-8')
            images_io = io.TextIOWrapper(zip_io.open("images.txt"), encoding='utf-8')

            while True:
                label = labels_io.readline().strip()
                feature = features_io.readline().strip()
                uuid_ = uuids_io.readline().strip()
                image_url = images_io.readline().strip()

                if label == '' or feature == '':
                    # stop the cycle if there are empty lines
                    break

                if not feature.startswith('[') and not feature.endswith(']'):
                    feature = f'[{feature}]'
                feature_vector = json.loads(feature)

                batch.append(DataEntry(
                    label,
                    feature_vector,
                    uuid.UUID(uuid_) if uuid_ else None,
                    image_url if image_url else None,
                ))
                # if len(batch) >= self.concurrency:
                if len(batch) >= self.concurrency:
                    yield batch
                    batch = []

            if batch:
                yield batch


def is_iterable(obj):
    try:
        iter(obj)
    except Exception:
        return False
    else:
        return True


class FileInputZipDatabasePacker(ZipDatabasePacker):
    """
    Class copied from rebotics_core and one to be used as main zipfile module database packer.
    """

    def pack(self, labels, features, *args, **kwargs):
        """
        Overload of the packer to utilize files instead of IO or numpy arrays
        """
        self.meta_data["count"] = kwargs.get("count")

        self.meta_data["additional_files"] = kwargs.get("additional_files", [])

        if (
            isinstance(labels, pathlib.PurePath)
            and isinstance(features, pathlib.PurePath)
            or isinstance(labels, str)
            and isinstance(features, str)
        ):  # check if it is a str path
            with self.zipfile(self.destination, mode="w") as zip_io:
                zip_io.write(str(labels), "labels.txt")
                zip_io.write(str(features), "features.txt")

                for filename, path in kwargs.get("additional_files", []):
                    zip_io.write(str(path), filename)

                zip_io.writestr("meta.json", json.dumps(self.meta_data))

        if is_iterable(labels) and is_iterable(features):
            return super(BaseZipClassificationDatabasePacker, self).pack(
                labels, features, *args, **kwargs
            )


class Py7zrClassificationDatabasePacker(BaseClassificationDatabasePacker):
    """
    Database packer with the py7zr implementation of pyzstd library.
    """
    allowed_path_types = (str, pathlib.PurePath)
    version = 3

    def __init__(
        self,
        archive_file: typing.Union[pathlib.PurePath, typing.AnyStr] = None,
        # check_duplicates=False,
        **kwargs
    ) -> None:
        """
        :param archive_file: string or path object of the archive file
        """
        if 'destination' in kwargs:
            destination = kwargs['destination']
            source = None
            del kwargs['destination']
        elif 'source' in kwargs:
            destination = None
            source = kwargs['source']
            del kwargs['source']
        else:
            destination = archive_file
            source = None

        super().__init__(source=source, destination=destination, **kwargs)
        self.meta_data['packer_version'] = self.version
        self.read_meta()

    @property
    def filters(self):
        return [{'id': py7zr.FILTER_DEFLATE, }]

    def archive_file(self, **kwargs) -> py7zr.SevenZipFile:
        # TODO: from created SevenZipFile object it is possible to create ArchiveInfo object, that contains all info
        #  about used filters, meaning it is possible to create filter-attribute-independent unpacker process
        return py7zr.SevenZipFile(self.destination, filters=self.filters, **kwargs)

    def pack(
        self,
        labels: Union[AnyStr, pathlib.PurePath],
        features: Union[AnyStr, pathlib.PurePath],
        *args, **kwargs
    ):
        """
        Packs provided label and feature files (by path) into the archive file. Additional arguments provided should
        correspond **kwargs of py7zr zip file interface.

        :param labels: label input file, path of some sort
        :param features: features input file, path of some sort
        """
        # TODO: additional file array, possibly
        file_state, error = validate_files([labels, features])
        if not file_state:
            raise FileNotFoundError(error)

        if isinstance(labels, str):
            labels = pathlib.Path(labels)
        if isinstance(features, str):
            features = pathlib.Path(features)

        archive_file = self.archive_file(mode='w', **kwargs)

        archive_file.write(labels, RCDBBaseFileNames.LABELS)
        archive_file.write(features, RCDBBaseFileNames.FEATURES)
        archive_file.writestr(json.dumps(self.meta_data), RCDBBaseFileNames.META)

        archive_file.close()

    def unpack(self):
        """
        Reads contents of RCDB archive into memory and provides generator of feature values.
        """
        file_state, error = validate_files([self.destination, ])
        if not file_state:
            raise FileNotFoundError(error)

        with self.archive_file(mode='r') as archive_file:
            data = archive_file.readall()

            if RCDBBaseFileNames.LABELS not in data or RCDBBaseFileNames.FEATURES not in data:
                raise UnpackException(archive_file.filename, list(data.keys()))

            labels = read_lines(data[RCDBBaseFileNames.LABELS].getvalue().decode('utf-8'), skip_empty_lines=True)
            features = read_lines(data[RCDBBaseFileNames.FEATURES].getvalue().decode('utf-8'), skip_empty_lines=True)

        for index, (label, feature) in enumerate(zip(labels, features)):
            yield ClassificationEntry(label, feature, None, None, index)

    def read_meta(self):
        if not os.path.exists(self.destination):
            return

        with self.archive_file(mode='r') as archive_file:
            data = archive_file.read(RCDBBaseFileNames.META)
            if RCDBBaseFileNames.META not in data:
                raise UnpackException(self.destination, list(data.keys()))

            self.meta_data = json.loads(data[RCDBBaseFileNames.META].getvalue().decode('utf-8'))



class ZstdClassificationDatabasePacker(Py7zrClassificationDatabasePacker):
    def __init__(
        self,
        archive_file: typing.Union[pathlib.PurePath, typing.AnyStr] = None,
        compression_level=7,
        **kwargs
    ):
        """
        :param archive_file: string or path object of the archive file
        :param compression_level: compression level in shared paradigm of py7zr and zstd, 1 to 22, from fastest to
        most compressed
        """
        self.__compression_level = compression_level
        super().__init__(archive_file, **kwargs)

    @property
    def compression_level(self):
        return self.__compression_level

    @property
    def filters(self):
        return [{'id': py7zr.FILTER_ZSTD, 'level': self.compression_level}]


@dataclasses.dataclass
class Packer:
    checker: typing.Callable
    # Basically, we're fine with any subclass of base class
    base_class: typing.Type[BaseClassificationDatabasePacker]


# THIS LIST IS ORDER-DEPENDENT
# Basically, zipfile.is_zipfile accepts 7z archives made by py7zr module. This is stupid, but
#
# it is how it is.
#
PACKERS: typing.List[Packer] = [
    Packer(checker=py7zr.is_7zfile, base_class=Py7zrClassificationDatabasePacker),
    Packer(checker=zipfile.is_zipfile, base_class=FileInputZipDatabasePacker),
]


def get_packer_class(file_path: typing.Union[AnyStr, pathlib.PurePath]) \
  -> typing.Union[typing.Type[BaseClassificationDatabasePacker], None]:
    if not os.path.exists(file_path):
        raise FileNotFoundError("Unable to identify database packer class for file that does not exist.")

    for packer in PACKERS:
        if packer.checker(file_path):
            return packer.base_class
    else:
        return None



import logging
import os
import pathlib
import sys
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from datetime import datetime
from math import inf

import click
import tqdm
from click import pass_context

from rebotics_sdk.cli.common import configure, roles, set_token, shell
from rebotics_sdk.cli.utils import (
    app_dir, pass_rebotics_context, process_role, read_saved_role, ReboticsCLIContext, ReboticsScriptsConfiguration,
)
from rebotics_sdk.providers.fvm import FVMProvider
from .. import utils
from ..advanced import remote_loaders
from ..advanced.packers import VirtualClassificationDatabasePacker
from ..providers import AdminProvider, ProviderHTTPClientException


@click.group()
@click.option('-f', '--format', default='table', type=click.Choice(['table', 'id', 'json']), help='Result rendering')
@click.option('-v', '--verbose', is_flag=True, help='Enables verbose mode')
@click.option('-c', '--config', type=click.Path(), default='fvm.json', help="Specify what config.json to use")
@click.option('-r', '--role', default=lambda: read_saved_role('fvm'), help="Key to specify what fvm to use")
@click.version_option()
@click.pass_context
def api(ctx, format, verbose, config, role):
    """
    Admin CLI tool to communicate with FVM API
    """
    process_role(ctx, role, 'fvm')
    ctx.obj = ReboticsCLIContext(
        'fvm',
        role,
        format,
        verbose,
        os.path.join(app_dir, config),
        provider_class=FVMProvider
    )


api.add_command(shell, 'shell')
api.add_command(roles, 'roles')
api.add_command(configure, 'configure')
api.add_command(set_token, 'set_token')


@api.command(name='file')
@click.option('-f', '--name', required=True, help='Filename of the file', type=click.UNPROCESSED)
@pass_rebotics_context
def virtual_upload(ctx, name):
    """Create virtual upload"""
    if ctx.verbose:
        click.echo('Calling create virtual upload')
    result = ctx.provider.create_virtual_upload(
        name
    )
    if 'id' in result.keys():
        pk = result['id']
        with open(name, 'rb', ) as fio:
            click.echo('Uploading file...')
            remote_loaders.upload(destination=result['destination'], file=fio, filename=name)
            ctx.provider.finish(
                pk
            )
            click.echo("Successfully finished uploading")
    else:
        click.echo("Failed to call virtual upload")


@api.group()
def rcdb():
    pass


def _download_rcdb_locally(ctx, rcdb_file, target):
    rcdb_file = str(rcdb_file)

    if rcdb_file.isdigit():
        # download rcdb file by ID
        try:
            response = ctx.provider.get_rcdb_by_id(int(rcdb_file))
        except ProviderHTTPClientException as exc:
            raise click.ClickException(f"Failed to trigger by API with error: {exc}")
        # should we also save a response? probably no
        return _download_rcdb_locally(ctx, response['file']['file'], target)  # get the url
    elif utils.is_url(rcdb_file):
        # download rcdb file by URL
        filename = utils.get_filename_from_url(rcdb_file)
        local_filepath = pathlib.Path(target / filename)
        if local_filepath.exists():
            return local_filepath
        try:
            remote_loaders.download(rcdb_file, local_filepath)
        except Exception as exc:
            raise click.ClickException(f"Failed to download file by URL: {exc}")
        return _download_rcdb_locally(ctx, local_filepath, target)
    else:
        # assuming that it is a local file path
        local_filepath = pathlib.Path(rcdb_file)
        if not local_filepath.exists():
            raise click.ClickException("Local file is not loaded!")
        return local_filepath


preview_downloader_dataclass_kwargs = {
    'frozen': True,
    'eq': False,
}
if sys.version_info >= (3, 10):
    preview_downloader_dataclass_kwargs['slots'] = True


@dataclass(**preview_downloader_dataclass_kwargs)
class PreviewDownloader:
    provider: FVMProvider
    progress_bar: tqdm.tqdm
    error_logger: logging.Logger

    def __call__(self, url: str, target: pathlib.Path) -> None:
        try:
            self.provider.download(url, target)
        except Exception as exc:
            self.error_logger.exception(f"Failed to download preview from {url} to {target}", exc_info=exc)
        self.progress_bar.update()


@rcdb.command(name="unpack")
@click.argument('rcdb_file')
@click.option('-t', '--target', type=click.Path(), default=pathlib.Path('.'))
@click.option('-w', '--with-images', is_flag=True)
@click.option('-c', '--concurrency', default=None, type=click.INT)
@pass_rebotics_context
def rcdb_unpack(ctx, rcdb_file, target, with_images, concurrency):
    """Unpack rcdb files. if the images """
    target = pathlib.Path(target)
    click.echo("Downloading rcdb file locally...")
    rcdb_file = _download_rcdb_locally(ctx, rcdb_file, target)
    if not with_images:
        click.echo(f"File loaded into {rcdb_file}")
        return

    error_logger = logging.getLogger('fvm.rcdb_unpack.errors')
    error_logger.setLevel(logging.ERROR)
    error_log_path = target / 'errors.log'
    error_logger_handler = logging.FileHandler(error_log_path, delay=True)
    error_logger_handler.setLevel(logging.ERROR)
    error_logger.addHandler(error_logger_handler)

    max_workers = concurrency
    if max_workers is not None:
        max_workers = max(max_workers, 1)

    features_path = target / 'features.txt'
    labels_path = target / 'labels.txt'
    images_path = target / 'images.txt'
    uuid_path = target / 'uuid.txt'

    # create new packer and unpacker
    packer = VirtualClassificationDatabasePacker(
        source=str(rcdb_file),
        with_images=with_images
    )
    features_count = packer.get_features_count()
    click.echo(f"Total features count: {features_count}")

    # Initialize downloading previews progress bar with certain params (leave, delay) to not display anything yet
    with tqdm.tqdm(desc="Downloading previews", leave=False, delay=inf) as download_previews_progress_bar, \
        ThreadPoolExecutor(max_workers=max_workers) as executor:  # noqa

        preview_downloader = PreviewDownloader(ctx.provider, download_previews_progress_bar, error_logger)
        total_images_to_download = 0

        with open(features_path, 'w') as features_io, \
            open(labels_path, 'w') as labels_io, \
            open(images_path, 'w') as images_io, \
            open(uuid_path, 'w') as uuid_io:  # noqa

            for entry in tqdm.tqdm(
                packer.unpack(), desc="Iterating over RCDB entries", total=features_count, leave=False, position=0
            ):
                # save images from entry
                image_path = target / entry.label / f"{entry.uuid}.jpeg"

                features_io.write(",".join(map(str, entry.feature)) + "\n")
                labels_io.write(entry.label + "\n")
                uuid_io.write(str(entry.uuid) + "\n")

                images_io.write(str(image_path) + "\n")
                if image_path.exists():
                    # check if image can be loaded via pil and whatnot
                    continue
                image_path.parent.mkdir(exist_ok=True, parents=True)

                # Enqueue a task for the executor to download image
                executor.submit(preview_downloader, entry.image_url, image_path)
                total_images_to_download += 1

        if total_images_to_download > 0:
            click.echo(f"Downloading {total_images_to_download} images not present in FS, please wait...")
            # Now we set params to display the progress bar, because there are previews to download
            download_previews_progress_bar.total = total_images_to_download
            download_previews_progress_bar.leave = True
            download_previews_progress_bar.delay = 0
            download_previews_progress_bar.refresh()

    if error_log_path.exists():
        click.echo(f"Some errors occurred, check {error_log_path} for details")

    click.echo("Unpacking is completed")


@rcdb.command(name='upload')
@click.argument('rcdb_file', type=click.Path(exists=True, dir_okay=False, file_okay=True))
@click.option('-r', '--retailer-codename', type=click.STRING)
@click.option('-m', '--model-codename', type=click.STRING)
@click.option('-w', '--with-images', is_flag=True)
@pass_rebotics_context
def rcdb_upload(ctx, rcdb_file, **kwargs):
    """
    Upload RCDB file
    """
    # retrieve model_id by the model_codename
    # read features count in the file

    rcdb_file = pathlib.Path(rcdb_file)
    kwargs['archive_size'] = rcdb_file.stat().st_size
    packer = VirtualClassificationDatabasePacker(source=str(rcdb_file))
    kwargs['features_count'] = packer.get_features_count()

    ctx.verbose_log(f"Archive size: {kwargs['archive_size']}. "
                    f"Features count: {kwargs['features_count']}")

    ctx.verbose_log("Upload rcdb file using presigned post URL")
    with open(rcdb_file, 'rb') as file_io:
        file_upload = ctx.provider.upload_file(file_io, filename=rcdb_file.name)
    file_id = file_upload['id']

    # call API to create an RCDB entry in database
    ctx.provider.save_rcdb(
        file_id,
        **kwargs
    )


@rcdb.command(name='copy')
@click.argument("backup_id", type=int)
@click.option("-r", "--retailer-codename", type=click.STRING, required=True, help="Retailer codename is required")
@click.option('-m', '--model-codename', type=click.STRING, help="If not specified,"
                                                                " the model_codename from the backup will be used")
@click.option("--admin-role", type=click.STRING, help="Uses the role, if not specified, "
                                                      "will be using the same role from fvm. "
                                                      "E.g. we have r3dev and r3us for both fvm and admin")
@pass_rebotics_context
@pass_context
def rcdb_copy_from_admin(click_ctx, ctx, backup_id, retailer_codename, model_codename, admin_role):
    """Copy CORE rcdb from the admin and upload and assign the rcdb file to appropriate retailer and model.

    To create rcdb backup from CORE go to admin application on /admin/nn_models/retailermodel/
    Click on "Download reference data"
    Wait a couple of minutes, and navigate to /admin/classification_data/featurevectorbackup/
    Retrieve the ID of the backup and use it with this command.
    """
    try:
        if admin_role is None:
            admin_role = ctx.role
            ctx.verbose_log(f"Admin role is not specified, using same role for fvm: {admin_role}")
        admin_config = ReboticsScriptsConfiguration(
            os.path.join(app_dir, 'admin.json'),
            provider_class=AdminProvider
        )
        admin_provider: AdminProvider = admin_config.get_provider(admin_role)

        if admin_provider is None:
            raise Exception(f"There is no admin provider with role name {admin_role}")
    except Exception as exc:
        click.echo(f"Failed to initialize admin provider with {exc}")
        sys.exit(1)

    ctx.verbose_log(f"Fetching rcdb backup {backup_id} from {admin_provider.host}")
    backup = admin_provider.rcdb.get(backup_id)
    ctx.verbose_log("Backup data: ")
    if ctx.verbose:
        ctx.format_result(backup, force_format='json')

    if backup['status'] != 'EXPORT_DONE':
        click.echo(f"The backup #{backup_id} export may not be finished and has status {backup['status']}")

    backup_url = backup['backup']
    rcdb_file_path = pathlib.Path(f"{retailer_codename}_{datetime.now().strftime('%Y_%m_%d_%H-%M')}.rcdb")

    ctx.verbose_log(f"Downloading rcdb backup from {backup_url} to {rcdb_file_path}")

    remote_loaders.download(backup_url, rcdb_file_path, progress_bar=True)

    ctx.verbose_log(f"Invoking rcdb_upload from {rcdb_file_path}")

    selected_model_codename = model_codename
    if model_codename is None:
        ctx.verbose_log('Setting model codename from backup')
        selected_model_codename = backup.get('model_codename')

    if selected_model_codename is None:
        click.echo("There is no specified model_codename.")
        sys.exit(1)

    click_ctx.invoke(rcdb_upload,
                     rcdb_file=rcdb_file_path,
                     retailer_codename=retailer_codename,
                     model_codename=selected_model_codename,
                     with_images=False)



import enum
import errno
import os
import pathlib
import re
from timeit import default_timer
from typing import List, AnyStr
from urllib.parse import uses_relative, uses_netloc, uses_params

import requests
from six.moves.urllib_parse import urlparse
from urllib3.util import parse_url

from rebotics_sdk.advanced import remote_loaders

_VALID_URLS = set(uses_relative + uses_netloc + uses_params)
_VALID_URLS.discard("")


def mkdir_p(path):
    try:
        os.makedirs(path)
    except OSError as exc:  # Python >2.5
        if exc.errno == errno.EEXIST and os.path.isdir(path):
            pass
        else:
            raise


def is_url(url) -> bool:
    """
    Check to see if a URL has a valid protocol.

    Parameters
    ----------
    url : str or unicode

    Returns
    -------
    isurl : bool
        If `url` has a valid protocol return True otherwise False.
    """
    if not isinstance(url, str):
        return False
    return parse_url(url).scheme in _VALID_URLS


def get_filename_from_url(url):
    return urlparse(url).path.split('/')[-1]


def download_file(url, filepath=None, **params):
    response = requests.get(
        url,
        stream=True,
        params=params,
        timeout=remote_loaders.HTTP_TIMEOUTS,
    )
    response.raise_for_status()

    if filepath is None:
        # TODO: decode url, get filename
        filepath = get_filename_from_url(url)

    with open(filepath, 'wb') as handle:
        for block in response.iter_content(1024):
            handle.write(block)
    return filepath


class Timer(object):
    def __init__(self):
        self.timer = default_timer

    def __enter__(self):
        self.start = self.timer()
        return self

    def __exit__(self, *args):
        end = self.timer()
        self.elapsed_secs = end - self.start
        self.elapsed = self.elapsed_secs * 1000  # millisecs


def parse_id_range_string(id_string):
    id_list = []
    range_pattern = re.compile(r'^(?P<from>\d+)-(?P<to>\d+)$')
    single_pattern = re.compile(r'^\d+$')

    left_out_parts = []

    for part in id_string.split(','):
        part = part.strip()
        range_match = range_pattern.match(part)

        if range_match:
            range_from = int(range_match.group('from'))
            range_to = int(range_match.group('to'))
            if range_from > range_to:
                left_out_parts.append(part)
                continue
            id_list.extend(range(range_from, range_to + 1))
        elif single_pattern.match(part):
            id_list.append(int(part))
        else:
            left_out_parts.append(part)

    if left_out_parts:
        raise ValueError('Please review the string as input: {}.\n Wrong parts are: {}'.format(
            id_string, left_out_parts
        ))
    if not id_list:
        raise ValueError('You need to supply a string with at least one ID. Given {}'.format(id_string))
    return sorted(list(set(id_list)))


def uri_validator(x):
    try:
        result = urlparse(x)
        return all([result.scheme, result.netloc, result.path])
    except Exception:
        return False


def download_and_return_id(row, output: pathlib.Path):
    url = row['url']
    label = row['label']
    uuid = row['uuid']

    label_folder = output / label
    label_folder.mkdir(exist_ok=True, parents=True)

    filename = pathlib.Path(urlparse(url).path).name
    destination = label_folder / filename

    retries = 4
    while retries != 0:
        try:
            remote_loaders.download(url, destination)
            return uuid, destination
        except Exception:
            retries -= 1

    return uuid, None


class FileState(enum.IntEnum):
    NOT_FOUND = 0
    EXISTS = 1


def validate_files(files: List) -> (FileState, str):
    error_message = ''
    for file in files:
        if not os.path.exists(file):
            error_message += f'Provided file \'{file}\' does not exist.\n'

    return FileState.NOT_FOUND if error_message else FileState.EXISTS, error_message


def read_lines(lines: AnyStr, skip_empty_lines=False) -> List[AnyStr]:
    return [line for line in lines.split('\n') if line or not skip_empty_lines]



# -*- coding: utf-8 -*-

"""Top-level package for rebotics_sdk."""

__author__ = """Malik Sulaimanov"""
__email__ = 'malik@retechlabs.com'
__version__ = '0.21.2'

try:
    from rebotics_sdk.cli.authenticated_provider import get_provider  # noqa: F401
except ImportError:
    pass



import json
import os
import pathlib
import tempfile
from unittest import TestCase

import pytest
import requests_mock

from rebotics_sdk.advanced.packers import (
    ClassificationDatabasePacker,
    DuplicateFeatureVectorsException,
    Py7zrClassificationDatabasePacker,
    FileInputZipDatabasePacker,
    VirtualClassificationDatabasePacker,
    ZipDatabasePacker,
    ZstdClassificationDatabasePacker,
    get_packer_class,
    PACKERS
)
from rebotics_sdk.constants import RCDBBaseFileNames
from rebotics_sdk.providers import FVMProvider


@pytest.fixture(scope="module")
def script_cwd(request):
    return request.fspath.join("..")


def test_classification_packing(script_cwd):
    db_folder = pathlib.Path(script_cwd.join("db"))
    with tempfile.TemporaryDirectory() as dirname:
        destination_filename = pathlib.Path(dirname, 'test')
        packer = ClassificationDatabasePacker(destination=destination_filename)
        features = db_folder / 'features.txt'
        labels = db_folder / 'labels.txt'
        images_folder = db_folder / 'custom_folder/'

        res = packer.pack(labels, features, images_folder)

        assert 'test.rcdb' in res

        with packer.zipfile(res, mode='r') as zip_io:
            meta = json.loads(packer.read_lines(zip_io.read('meta.json').decode('utf-8'))[0])

        expected_keys = [
            'packed',
            'model_type',
            'model_codename',
            'sdk_version',
            'packer_version',
            'core_version',
            'fvm_version'
        ]
        for key in expected_keys:
            assert key in meta.keys()

        assert len(packer.images) == 2

        packer = ClassificationDatabasePacker(source=res)
        entries = list(packer.unpack())
        assert len(entries) == 2
        entry = entries[0]
        assert entry.label == '123123123'
        assert entry.feature == '123123123123123'
        internal_filename = entry.filename
        assert internal_filename == 'image_1.png'

        for key in expected_keys:
            assert key in packer.meta_data.keys()

        # testing if it can be dumped to the FS
        og_file = images_folder / internal_filename
        tmp_file = db_folder / internal_filename

        with open(tmp_file, 'wb') as fout:
            fout.write(entry.image)

        assert og_file.stat().st_size == tmp_file.stat().st_size
        tmp_file.unlink()


def test_classification_packing_check_duplicates(script_cwd):
    db_folder = pathlib.Path(script_cwd.join("db"))
    packer = ClassificationDatabasePacker(destination='test', check_duplicates=True)
    features = db_folder / 'features.txt'
    labels = db_folder / 'labels.txt'
    images_folder = db_folder / 'custom_folder/'

    with pytest.raises(DuplicateFeatureVectorsException) as excinfo:
        packer.pack(labels, features, images_folder)
    assert "duplicate" in str(excinfo.value)


def test_zip_packing():
    packer = ZipDatabasePacker()
    packed = packer.pack(
        labels=[
            '123123123'
        ],
        features=[
            '123123123123123'
        ]
    )
    assert packer.meta_data['count'] == 1

    unpacker = ZipDatabasePacker(source=packed)
    for entry in unpacker.unpack():
        assert entry.label == '123123123'
        assert entry.feature == '123123123123123'

    assert unpacker.meta_data['count'] == 1


def test_virtual_packing_and_unpacking(script_cwd):
    db_folder = pathlib.Path(script_cwd.join("db"))

    with tempfile.TemporaryDirectory() as dirname, requests_mock.Mocker() as m:
        m.get('https://via.placeholder.com/150', text='some file')
        destination_filename = pathlib.Path(dirname, 'test')
        # destination_filename = db_folder / "test.rcdb"
        provider = FVMProvider(host='https://r3dev-fvm.rebotics.net/')
        packer = VirtualClassificationDatabasePacker(
            destination=destination_filename,
            provider=provider,
        )

        features = db_folder / 'features.txt'
        labels = db_folder / 'labels.txt'
        images = db_folder / 'image_urls.txt'
        uuids = db_folder / 'uuid.txt'

        res = packer.pack(
            labels, features, uuids, images
        )
        assert 'test.rcdb' in res, "Same destination is returned properly and extension is set normally"

        unpacker = VirtualClassificationDatabasePacker(
            source=pathlib.Path(dirname, 'test.rcdb'),
            with_images=True,
            provider=provider,
        )
        data = list(unpacker.unpack())
        assert len(data) == 2, "There are only two entries along the way"


def test_different_comrpessionlevel_compatibility(script_cwd):
    db_folder = pathlib.Path(script_cwd).joinpath('db')

    with open(db_folder.joinpath(RCDBBaseFileNames.LABELS)) as labels_file:
        labels = labels_file.read().splitlines()

    with open(db_folder.joinpath(RCDBBaseFileNames.FEATURES)) as features_file:
        features = features_file.read().splitlines()

    with tempfile.TemporaryDirectory() as dirname:
        archive_file_path = pathlib.Path(dirname).joinpath('fv.zip')

        database_packer = ZipDatabasePacker(destination=archive_file_path, compresslevel=3)
        database_packer.pack(labels, features)

        database_unpacker = ZipDatabasePacker(source=archive_file_path)
        data = list(database_unpacker.unpack())

        assert len(data) == 2


@pytest.mark.skip
def test_full_data():
    """
    This is a custom test that requires preparation and must be skipped by default.
    """
    import datetime as dt
    db_source = pathlib.Path('G:\\retech\\rebotics_core\\docker-files\\test-archives\\fv')
    archive_name = 'fv.zip'
    labels_path = pathlib.Path(db_source).joinpath(RCDBBaseFileNames.LABELS)
    features_path = pathlib.Path(db_source).joinpath(RCDBBaseFileNames.FEATURES)
    log = []

    def run_compression_program(temporary_directory, compression_level):
        archive_path = temporary_directory.joinpath(archive_name)
        packer = FileInputZipDatabasePacker(destination=archive_path, compresslevel=compression_level)
        pack_start = dt.datetime.now()
        packer.pack(labels_path, features_path)
        pack_end = dt.datetime.now()
        del packer

        unpacker = FileInputZipDatabasePacker(source=archive_path)
        unpack_start = dt.datetime.now()
        data = list(unpacker.unpack())
        unpack_end = dt.datetime.now()
        del unpacker

        return {
            'compression_level': compression_level,
            # 'pack_start': pack_start,
            # 'pack_end': pack_end,
            'pack_duration': str(pack_end - pack_start),
            # 'unpack_start': unpack_start,
            # 'unpack_end': unpack_end,
            'unpack_duration': str(unpack_end - unpack_start),
        }

    with tempfile.TemporaryDirectory() as dir_c1, tempfile.TemporaryDirectory() as dir_c6:
        for folder, compresslevel in ((pathlib.Path(dir_c1), 1), (pathlib.Path(dir_c6), 6)):
            result = run_compression_program(folder, compresslevel)
            log.append(result)

    print(json.dumps(log, indent=4))


class Py7zrClassificationDatabasePackerTestCase(TestCase):
    archiver = Py7zrClassificationDatabasePacker
    archive_packed_flag = False
    magic_text = 'Lorem ipsum dolor sit amet.'

    def test_archiver(self):
        local_path = pathlib.Path(os.path.realpath(__file__)).parent
        self.database_source = local_path.joinpath('db')

        with tempfile.TemporaryDirectory() as working_directory:
            self.archive_file = pathlib.Path(working_directory).joinpath('fv')\
                .with_suffix('.' + self.archiver.extension)

            with self.subTest("Archive file packing test"):
                self.pack_database()

            with self.subTest("Archive file unpacking test"):
                self.unpack_database()

    def pack_database(self):
        self.archive_handler = self.archiver(self.archive_file)
        self.archive_handler.meta_data['test_easter_egg'] = self.magic_text
        self.archive_handler.pack(
            self.database_source.joinpath(RCDBBaseFileNames.LABELS),
            self.database_source.joinpath(RCDBBaseFileNames.FEATURES)
        )

        self.assertTrue(os.path.exists(self.archive_file))
        self.archive_packed_flag = True

    def unpack_database(self):
        if not self.archive_packed_flag:
            self.skipTest("Failed to create archive in previous test method "
                          "Py7zrClassificationDatabasePackerTestCase.pack_database")

        data = list(self.archive_handler.unpack())
        self.assertTrue(len(data) == 2)
        self.assertTrue(self.archive_handler.meta_data['packer_version'] == 3)
        self.assertTrue(self.archive_handler.meta_data['test_easter_egg'] == 'Lorem ipsum dolor sit amet.')


class ZstdClassificationDatabasePackerTestCase(Py7zrClassificationDatabasePackerTestCase):
    archiver = ZstdClassificationDatabasePacker


class GetPackerTestCase(TestCase):
    archive_name = 'fv'

    def setUp(self) -> None:
        self.test_data_path = pathlib.Path(os.path.realpath(__file__)).parent.joinpath('db')

    def test_packers(self):
        for packer in PACKERS:
            with tempfile.TemporaryDirectory() as folder:
                archive_path = pathlib.Path(folder).joinpath(self.archive_name)\
                    .with_suffix('.' + packer.base_class.extension)

                archive = packer.base_class(destination=archive_path)
                archive.pack(
                    self.test_data_path.joinpath(RCDBBaseFileNames.LABELS),
                    self.test_data_path.joinpath(RCDBBaseFileNames.FEATURES)
                )
                del archive

                type_derived = get_packer_class(archive_path)

                self.assertTrue(type_derived is packer.base_class)
